/**
 * Copyright (c) 2022, Salesforce, Inc.,
 * All rights reserved.
 * For full license text, see the LICENSE.txt file
 */

/*
 *  ATTENTION!
 *  THIS IS A GENERATED FILE FROM https://github.com/salesforce-experience-platform-emu/lds-lightning-platform
 *  If you would like to contribute to LDS, please follow the steps outlined in the git repo.
 *  Any changes made to this file in p4 will be automatically overwritten.
 *  *******************************************************************************************
 */
import { setDefaultLuvio } from 'force/ldsEngine';
import { setBypassDeepFreeze, StoreKeySet, serializeStructuredKey, StringKeyInMemoryStore, Reader, deepFreeze, emitAdapterEvent, InMemoryStore, Environment, Luvio } from 'force/luvioEngine';
import { instrumentLuvio } from 'force/ldsInstrumentation';
import { isStoreKeyRecordViewEntity, RECORD_ID_PREFIX, RECORD_FIELDS_KEY_JUNCTION, getRecordId18, extractRecordIdFromStoreKey, RECORD_VIEW_ENTITY_ID_PREFIX, keyBuilderRecord } from 'force/ldsAdaptersUiapi';
import '@salesforce/gate/lds.idempotencyWriteDisabled';
import '@salesforce/gate/lds.backdatingEnabled';
import { Kind, buildSchema, isObjectType, defaultFieldResolver } from 'force/ldsGraphqlParser';
import FIRST_DAY_OF_WEEK from '@salesforce/i18n/firstDayOfWeek';
import excludeStaleRecordsGate from '@salesforce/gate/lds.graphqlEvalExcludeStaleRecords';
import networkAdapter from 'force/ldsNetwork';
import ldsEngineCreator from 'force/ldsEngineCreator';

// the last version the metadata shape was altered
const DURABLE_METADATA_VERSION = '0.111.0';
function isDeprecatedDurableStoreEntry(durableRecord) {
    if (durableRecord.expiration !== undefined) {
        return true;
    }
    const metadata = durableRecord.metadata;
    if (metadata !== undefined) {
        const { metadataVersion } = metadata;
        // eventually we will want to assert that metadataVersion is defined
        if (metadataVersion !== undefined && metadataVersion !== DURABLE_METADATA_VERSION) {
            return true;
        }
    }
    // Add more deprecated shape checks here
    return false;
}
const DefaultDurableSegment = 'DEFAULT';
const RedirectDurableSegment = 'REDIRECT_KEYS';
const MessagingDurableSegment = 'MESSAGING';
const MessageNotifyStoreUpdateAvailable = 'notifyStoreUpdateAvailable';

const { keys: keys$3, create: create$3, assign: assign$3, freeze: freeze$1 } = Object;

//Durable store error instrumentation key
const DURABLE_STORE_ERROR = 'durable-store-error';
/**
 * Returns a function that processes errors from durable store promise rejections.
 * If running in a non-production environment, the error is rethrown.
 * When running in production the error is sent to instrumentation.
 * @param instrument Instrumentation function implementation
 */
function handleDurableStoreRejection(instrument) {
    return (error) => {
        if (process.env.NODE_ENV !== 'production') {
            throw error;
        }
        if (instrument !== undefined) {
            instrument(() => {
                return {
                    [DURABLE_STORE_ERROR]: true,
                    error: error,
                };
            });
        }
    };
}

function isStoreEntryError(storeRecord) {
    if (!storeRecord || typeof storeRecord !== 'object') {
        return false;
    }
    return storeRecord.__type === 'error';
}

/**
 * Takes a set of entries from DurableStore and publishes them via the passed in funcs.
 * This respects expiration and checks for valid DurableStore data shapes.  This should
 * be used over manually parsing DurableStoreEntries
 *
 * @param durableRecords The DurableStoreEntries to parse
 * @param publish A function to call with the data of each DurableStoreEntry
 * @param publishMetadata A function to call with the metadata of each DurableStoreEntry
 * @param pendingWriter the PendingWriter (this is going away soon)
 * @returns
 */
function publishDurableStoreEntries(durableRecords, put, publishMetadata) {
    const revivedKeys = new StoreKeySet();
    let hadUnexpectedShape = false;
    if (durableRecords === undefined) {
        return { revivedKeys, hadUnexpectedShape };
    }
    const durableKeys = keys$3(durableRecords);
    if (durableKeys.length === 0) {
        // no records to revive
        return { revivedKeys, hadUnexpectedShape };
    }
    for (let i = 0, len = durableKeys.length; i < len; i += 1) {
        const key = durableKeys[i];
        const durableRecord = durableRecords[key];
        if (isDeprecatedDurableStoreEntry(durableRecord)) {
            // had the old shape, skip reviving this entry.
            hadUnexpectedShape = true;
            continue;
        }
        const { metadata, data } = durableRecord;
        if (data === undefined) {
            // if unexpected data skip reviving
            hadUnexpectedShape = true;
            continue;
        }
        if (metadata !== undefined) {
            const { expirationTimestamp } = metadata;
            if (expirationTimestamp === undefined) {
                // if unexpected expiration data skip reviving
                hadUnexpectedShape = true;
                continue;
            }
            publishMetadata(key, metadata);
        }
        if (isStoreEntryError(data)) {
            // freeze errors on way into L1
            deepFreeze(data.error);
        }
        put(key, data);
        revivedKeys.add(key);
    }
    return { revivedKeys, hadUnexpectedShape };
}
/**
 * This method returns a Promise to a snapshot that is revived from L2 cache.  If
 * L2 does not have the entries necessary to fulfill the snapshot then this method
 * will refresh the snapshot from network, and then run the results from network
 * through L2 ingestion, returning the subsequent revived snapshot.
 */
function reviveSnapshot(baseEnvironment, durableStore, unavailableSnapshot, durableStoreErrorHandler, buildL1Snapshot, revivingStore, reviveMetrics = { l2Trips: [] }) {
    const { recordId, select, missingLinks, seenRecords, state } = unavailableSnapshot;
    // L2 can only revive Unfulfilled snapshots that have a selector since they have the
    // info needed to revive (like missingLinks) and rebuild.  Otherwise return L1 snapshot.
    if (state !== 'Unfulfilled' || select === undefined) {
        return Promise.resolve({
            snapshot: unavailableSnapshot,
            metrics: reviveMetrics,
        });
    }
    const keysToReviveSet = new StoreKeySet();
    if (revivingStore) {
        // Any stale keys since the last l2 read should be cleared and fetched again
        for (const staleKey of revivingStore.staleEntries) {
            keysToReviveSet.add(staleKey);
        }
        revivingStore.clearStale();
    }
    else {
        // when not using a reviving store:
        // in case L1 store changes/deallocs a record while we are doing the async read
        // we attempt to read all keys from L2 - so combine recordId with any seenRecords
        keysToReviveSet.add(recordId);
        keysToReviveSet.merge(seenRecords);
    }
    keysToReviveSet.merge(missingLinks);
    const keysToRevive = keysToReviveSet.keysAsArray();
    const canonicalKeys = keysToRevive.map((x) => serializeStructuredKey(baseEnvironment.storeGetCanonicalKey(x)));
    const start = Date.now();
    const { l2Trips } = reviveMetrics;
    return durableStore.getEntries(canonicalKeys, DefaultDurableSegment).then((durableRecords) => {
        l2Trips.push({
            duration: Date.now() - start,
            keysRequestedCount: canonicalKeys.length,
        });
        const { revivedKeys, hadUnexpectedShape } = publishDurableStoreEntries(durableRecords, 
        // TODO [W-10072584]: instead of implicitly using L1 we should take in
        // publish and publishMetadata funcs, so callers can decide where to
        // revive to (like they pass in how to do the buildL1Snapshot)
        baseEnvironment.storePut.bind(baseEnvironment), baseEnvironment.publishStoreMetadata.bind(baseEnvironment));
        // if the data coming back from DS had an unexpected shape then just
        // return the L1 snapshot
        if (hadUnexpectedShape === true) {
            return { snapshot: unavailableSnapshot, metrics: reviveMetrics };
        }
        if (revivedKeys.size() === 0) {
            // durable store doesn't have what we asked for so return L1 snapshot
            return { snapshot: unavailableSnapshot, metrics: reviveMetrics };
        }
        // try building the snapshot from L1 now that we have revived the missingLinks
        const snapshot = buildL1Snapshot();
        // if snapshot is pending then some other in-flight refresh will broadcast
        // later
        if (snapshot.state === 'Pending') {
            return { snapshot, metrics: reviveMetrics };
        }
        if (snapshot.state === 'Unfulfilled') {
            // have to check if the new snapshot has any additional seenRecords
            // and revive again if so
            const { seenRecords: newSnapshotSeenRecords, missingLinks: newSnapshotMissingLinks, recordId: newSnapshotRecordId, } = snapshot;
            const newKeysToReviveSet = new StoreKeySet();
            newKeysToReviveSet.add(newSnapshotRecordId);
            newKeysToReviveSet.merge(newSnapshotSeenRecords);
            newKeysToReviveSet.merge(newSnapshotMissingLinks);
            const newKeys = newKeysToReviveSet.keysAsArray();
            // in case DS returned additional entries we combine the requested
            // and returned keys
            const alreadyRequestedOrRevivedSet = keysToReviveSet;
            alreadyRequestedOrRevivedSet.merge(revivedKeys);
            // if there's any seen keys in the newly rebuilt snapshot that
            // haven't already been requested or returned then revive again
            for (let i = 0, len = newKeys.length; i < len; i++) {
                const newSnapshotSeenKey = newKeys[i];
                if (!alreadyRequestedOrRevivedSet.has(newSnapshotSeenKey)) {
                    return reviveSnapshot(baseEnvironment, durableStore, snapshot, durableStoreErrorHandler, buildL1Snapshot, revivingStore, reviveMetrics);
                }
            }
        }
        return { snapshot, metrics: reviveMetrics };
    }, (error) => {
        durableStoreErrorHandler(error);
        // getEntries failed, return the L1 snapshot
        return { snapshot: unavailableSnapshot, metrics: reviveMetrics };
    });
}

const TTL_DURABLE_SEGMENT = 'TTL_DURABLE_SEGMENT';
const TTL_DEFAULT_KEY = 'TTL_DEFAULT_KEY';
function buildDurableTTLOverrideStoreKey(namespace, representationName) {
    return `${namespace}::${representationName}`;
}
function isEntryDurableTTLOverride(entry) {
    if (typeof entry === 'object' && entry !== undefined && entry !== null) {
        const data = entry.data;
        if (data !== undefined) {
            return (data.namespace !== undefined &&
                data.representationName !== undefined &&
                data.ttl !== undefined);
        }
    }
    return false;
}
function isDefaultDurableTTLOverride(override) {
    return (override.namespace === TTL_DEFAULT_KEY && override.representationName === TTL_DEFAULT_KEY);
}
/**
 * Class to set and get the TTL override values in the Durable Store
 */
class DurableTTLStore {
    constructor(durableStore) {
        this.durableStore = durableStore;
    }
    setDefaultDurableTTLOverrides(ttl) {
        return this.durableStore.setEntries({
            [buildDurableTTLOverrideStoreKey(TTL_DEFAULT_KEY, TTL_DEFAULT_KEY)]: {
                data: {
                    namespace: TTL_DEFAULT_KEY,
                    representationName: TTL_DEFAULT_KEY,
                    ttl,
                },
            },
        }, TTL_DURABLE_SEGMENT);
    }
    setDurableTTLOverride(namespace, representationName, ttl) {
        return this.durableStore.setEntries({
            [buildDurableTTLOverrideStoreKey(namespace, representationName)]: {
                data: { namespace, representationName, ttl },
            },
        }, TTL_DURABLE_SEGMENT);
    }
    getDurableTTLOverrides() {
        return this.durableStore
            .getAllEntries(TTL_DURABLE_SEGMENT)
            .then((entries) => {
            const overrides = [];
            let defaultTTL = undefined;
            if (entries === undefined) {
                return {
                    defaultTTL,
                    overrides,
                };
            }
            const keys$1 = keys$3(entries);
            for (let i = 0, len = keys$1.length; i < len; i++) {
                const key = keys$1[i];
                const entry = entries[key];
                if (entry !== undefined && isEntryDurableTTLOverride(entry)) {
                    if (isDefaultDurableTTLOverride(entry.data)) {
                        defaultTTL = entry.data;
                    }
                    else {
                        overrides.push(entry.data);
                    }
                }
            }
            return {
                defaultTTL,
                overrides,
            };
        });
    }
}

function flushInMemoryStoreValuesToDurableStore(store, durableStore, durableStoreErrorHandler, redirects, additionalDurableStoreOperations = [], enableDurableMetadataRefresh = false) {
    const durableRecords = create$3(null);
    const refreshedDurableRecords = create$3(null);
    const evictedRecords = create$3(null);
    const { visitedIds, refreshedIds } = store.fallbackStringKeyInMemoryStore;
    // TODO: W-8909393 Once metadata is stored in its own segment we need to
    // call setEntries for the visitedIds on default segment and call setEntries
    // on the metadata segment for the refreshedIds
    const keys$1 = keys$3({ ...visitedIds, ...refreshedIds });
    for (let i = 0, len = keys$1.length; i < len; i += 1) {
        const key = keys$1[i];
        const canonicalKey = store.getCanonicalRecordId(key);
        // this record has been redirected, evict the original
        if (canonicalKey !== key) {
            evictedRecords[key] = true;
            continue;
        }
        const record = store.readEntry(key);
        const wasVisited = visitedIds[key] !== undefined;
        // this record has been evicted, evict from DS
        if (wasVisited && record === undefined) {
            evictedRecords[key] = true;
            continue;
        }
        const metadata = store.readMetadata(key);
        const entries = wasVisited === true || enableDurableMetadataRefresh === false
            ? durableRecords
            : refreshedDurableRecords;
        setRecordTo(entries, key, record, metadata);
    }
    const durableStoreOperations = additionalDurableStoreOperations;
    const recordKeys = keys$3(durableRecords);
    if (recordKeys.length > 0) {
        // publishes with data
        durableStoreOperations.push({
            type: 'setEntries',
            entries: durableRecords,
            segment: DefaultDurableSegment,
        });
    }
    const refreshKeys = keys$3(refreshedDurableRecords);
    if (refreshKeys.length > 0) {
        // publishes with only metadata updates
        durableStoreOperations.push({
            type: 'setMetadata',
            entries: refreshedDurableRecords,
            segment: DefaultDurableSegment,
        });
    }
    // redirects
    redirects.forEach((value, key) => {
        durableStoreOperations.push({
            type: 'setEntries',
            entries: {
                [key]: {
                    data: { key, redirect: value },
                },
            },
            segment: RedirectDurableSegment,
        });
    });
    // evicts
    const evictedKeys = keys$3(evictedRecords);
    if (evictedKeys.length > 0) {
        durableStoreOperations.push({
            type: 'evictEntries',
            ids: evictedKeys,
            segment: DefaultDurableSegment,
        });
    }
    if (durableStoreOperations.length > 0) {
        return durableStore.batchOperations(durableStoreOperations).catch(durableStoreErrorHandler);
    }
    return Promise.resolve();
}
function setRecordTo(entries, key, record, metadata) {
    entries[key] = {
        data: record,
    };
    if (metadata !== undefined) {
        entries[key].metadata = {
            ...metadata,
            metadataVersion: DURABLE_METADATA_VERSION,
        };
    }
}

const DurableEnvironmentEventDiscriminator = 'durable';
function emitDurableEnvironmentAdapterEvent(eventData, observers) {
    emitAdapterEvent({
        type: 'environment',
        timestamp: Date.now(),
        environment: DurableEnvironmentEventDiscriminator,
        data: eventData,
    }, observers);
}

async function reviveTTLOverrides(ttlStore, environment) {
    const map = await ttlStore.getDurableTTLOverrides();
    const { defaultTTL, overrides } = map;
    if (defaultTTL !== undefined) {
        environment.storeSetDefaultTTLOverride(defaultTTL.ttl);
    }
    for (let i = 0, len = overrides.length; i < len; i++) {
        const { namespace, representationName, ttl } = overrides[i];
        environment.storeSetTTLOverride(namespace, representationName, ttl);
    }
}

/**
 * Returns an empty InMemoryStore that can be used for ingestion.  Copies over
 * the TTLOverrides from the given Environment's Store.
 */
function buildIngestStagingStore(environment) {
    return environment.storeBuildIngestionStagingStore();
}

async function reviveRedirects(durableStore, env) {
    const entries = await durableStore.getAllEntries(RedirectDurableSegment);
    if (entries) {
        for (const durableEntry of Object.keys(entries)) {
            const entry = entries[durableEntry];
            const { data: { key, redirect }, } = entry;
            if (entry) {
                env.storeRedirect(key, redirect);
            }
        }
    }
}

function buildRevivingStagingStore(upstreamStore) {
    const localStore = new StringKeyInMemoryStore();
    const staleEntries = new Set();
    function readEntry(key) {
        if (typeof key !== 'string') {
            return upstreamStore.readEntry(key);
        }
        let storeEntry = localStore.readEntry(key);
        if (!storeEntry) {
            // read from upstream store...
            storeEntry = upstreamStore.readEntry(key);
            // put it in our store to avoid it getting evicted prior to the next durable store read
            localStore.put(key, storeEntry);
        }
        return storeEntry;
    }
    // Entries are marked stale by the durable store change listener. They are not
    // immediately evicted so as to not result in a cache miss during a rebuild.
    // The revive process will clear stale entries and read them from the durable store
    // on the next revive loop.
    function markStale(key) {
        staleEntries.add(key);
    }
    // The revive loop clears stale entries right before reading from the durable store.
    // Any stale entries will be revived to ensure they are present in L1 and match the
    // latest data.
    function clearStale() {
        for (const key of staleEntries) {
            localStore.dealloc(key);
        }
        staleEntries.clear();
    }
    // All functions other than `readEntry` pass through to the upstream store.
    // A reviving store is only "active" during a call to `environment.storeLookup`, and will
    // be used by the reader attempting to build an L1 snapshot. Immediately after the L1 rebuild
    // the reviving store becomes inactive other than receiving change notifications.
    return create$3(upstreamStore, {
        readEntry: { value: readEntry },
        markStale: { value: markStale },
        clearStale: { value: clearStale },
        staleEntries: { value: staleEntries },
    });
}

const AdapterContextSegment = 'ADAPTER-CONTEXT';
const ADAPTER_CONTEXT_ID_SUFFIX = '__NAMED_CONTEXT';
async function reviveOrCreateContext(adapterId, durableStore, durableStoreErrorHandler, contextStores, pendingContextStoreKeys, onContextLoaded) {
    // initialize empty context store
    contextStores[adapterId] = create$3(null);
    const context = {
        set(key, value) {
            contextStores[adapterId][key] = value;
            durableStore.setEntries({
                [adapterId]: { data: contextStores[adapterId] },
            }, AdapterContextSegment);
            pendingContextStoreKeys.add(adapterId);
        },
        get(key) {
            return contextStores[adapterId][key];
        },
    };
    const contextReturn = () => {
        if (onContextLoaded !== undefined) {
            return onContextLoaded(context).then(() => {
                return context;
            });
        }
        return context;
    };
    try {
        const entries = await durableStore.getEntries([adapterId], AdapterContextSegment);
        if (entries !== undefined && entries[adapterId] !== undefined) {
            // if durable store has a saved context then load it in the store
            contextStores[adapterId] = entries[adapterId].data;
        }
    }
    catch (error) {
        durableStoreErrorHandler(error);
    }
    return contextReturn();
}
function isUnfulfilledSnapshot(cachedSnapshotResult) {
    if (cachedSnapshotResult === undefined) {
        return false;
    }
    if ('then' in cachedSnapshotResult) {
        return false;
    }
    return cachedSnapshotResult.state === 'Unfulfilled';
}
/**
 * Configures the environment to persist data into a durable store and attempt to resolve
 * data from the persistent store before hitting the network.  Sets the default
 * cache policy to stale-while-revalidate with infinite staleDuration.
 *
 * @param environment The base environment
 * @param durableStore A DurableStore implementation
 * @param instrumentation An instrumentation function implementation
 */
function makeDurable(environment, { durableStore, instrumentation, useRevivingStore, enableDurableMetadataRefresh = false, disableDeepFreeze = false, }) {
    // runtimes can choose to disable deepFreeze, e.g. headless mobile runtime
    setBypassDeepFreeze(disableDeepFreeze);
    let stagingStore = null;
    const durableTTLStore = new DurableTTLStore(durableStore);
    const mergeKeysPromiseMap = new Map();
    // When a context store is mutated we write it to L2, which causes DS on change
    // event. If this instance of makeDurable caused that L2 write we can ignore that
    // on change event. This Set helps us do that.
    const pendingContextStoreKeys = new Set();
    // Reviving stores are tracked so that they can be notified of durable store change notifications.
    const revivingStores = new Set();
    // redirects that need to be flushed to the durable store
    const pendingStoreRedirects = new Map();
    const contextStores = create$3(null);
    let initializationPromise = new Promise((resolve) => {
        const finish = () => {
            resolve();
            initializationPromise = undefined;
        };
        Promise.all([
            reviveTTLOverrides(durableTTLStore, environment),
            reviveRedirects(durableStore, environment),
        ]).then(finish);
    });
    //instrumentation for durable store errors
    const durableStoreErrorHandler = handleDurableStoreRejection(instrumentation);
    let disposed = false;
    const validateNotDisposed = () => {
        if (disposed === true) {
            throw new Error('This makeDurable instance has been disposed');
        }
    };
    const unsubscribe = durableStore.registerOnChangedListener(async (changes) => {
        const defaultSegmentKeys = [];
        const adapterContextSegmentKeys = [];
        const redirectSegmentKeys = [];
        const metadataRefreshSegmentKeys = [];
        const messagingSegmentKeys = [];
        let shouldBroadcast = false;
        for (let i = 0, len = changes.length; i < len; i++) {
            const change = changes[i];
            // we only care about changes to the data which is stored in the default
            // segment or the adapter context
            if (change.segment === DefaultDurableSegment) {
                if (change.type === 'setMetadata') {
                    metadataRefreshSegmentKeys.push(...change.ids);
                }
                else {
                    defaultSegmentKeys.push(...change.ids);
                }
            }
            else if (change.segment === AdapterContextSegment) {
                adapterContextSegmentKeys.push(...change.ids);
            }
            else if (change.segment === RedirectDurableSegment) {
                redirectSegmentKeys.push(...change.ids);
            }
            else if (change.segment === MessagingDurableSegment) {
                messagingSegmentKeys.push(...change.ids);
            }
        }
        if (redirectSegmentKeys.length > 0) {
            const redirectEntries = await durableStore.getEntries(redirectSegmentKeys, RedirectDurableSegment);
            if (redirectEntries !== undefined) {
                const redirectKeys = Object.keys(redirectEntries);
                for (const key of redirectKeys) {
                    const redirectData = redirectEntries[key];
                    environment.storeRedirect(redirectData.data.key, redirectData.data.redirect);
                    shouldBroadcast = true;
                }
            }
        }
        // process adapter context changes
        const adapterContextKeysFromDifferentInstance = [];
        for (const key of adapterContextSegmentKeys) {
            if (pendingContextStoreKeys.has(key)) {
                // if this instance caused the L2 write then remove from the
                // "pending" Set and move on
                pendingContextStoreKeys.delete(key);
            }
            else {
                // else it came from another luvio instance and we need to
                // read from L2
                adapterContextKeysFromDifferentInstance.push(key);
            }
        }
        if (adapterContextKeysFromDifferentInstance.length > 0) {
            try {
                const entries = await durableStore.getEntries(adapterContextKeysFromDifferentInstance, AdapterContextSegment);
                if (entries !== undefined) {
                    const entryKeys = keys$3(entries);
                    for (let i = 0, len = entryKeys.length; i < len; i++) {
                        const entryKey = entryKeys[i];
                        const entry = entries[entryKey];
                        contextStores[entryKey] = entry.data;
                    }
                }
            }
            catch (error) {
                durableStoreErrorHandler(error);
            }
        }
        // process default segment changes
        const defaultSegmentKeysLength = defaultSegmentKeys.length;
        if (defaultSegmentKeysLength > 0) {
            for (let i = 0; i < defaultSegmentKeysLength; i++) {
                const key = defaultSegmentKeys[i];
                // TODO: W-8909393 If expiration is the only thing that changed we should not evict the data... so
                // if we stored expiration and data at different keys (or same keys in different segments)
                // then we could know if only the expiration has changed and we wouldn't need to evict
                // and go through an entire broadcast/revive cycle for unchanged data
                // call base environment storeEvict so this evict is not tracked for durable deletion
                environment.storeEvict(key);
                for (const revivingStore of revivingStores) {
                    revivingStore.markStale(key);
                }
            }
            shouldBroadcast = true;
        }
        // process metadata only refreshes
        if (metadataRefreshSegmentKeys.length > 0) {
            const filteredKeys = metadataRefreshSegmentKeys.filter((s) => environment.storeKeyExists(s));
            if (filteredKeys.length > 0) {
                const entries = await durableStore.getMetadata(filteredKeys, DefaultDurableSegment);
                if (entries !== undefined) {
                    const entryKeys = keys$3(entries);
                    for (let i = 0, len = entryKeys.length; i < len; i++) {
                        const entryKey = entryKeys[i];
                        const { metadata } = entries[entryKey];
                        if (metadata !== undefined) {
                            environment.putStoreMetadata(entryKey, metadata, false);
                        }
                    }
                }
            }
        }
        if (shouldBroadcast) {
            await environment.storeBroadcast(rebuildSnapshot, environment.snapshotAvailable);
        }
        // if having notifyStoreUpdateAvailable msg, then pull the message body out from store
        // and call environment.notifyStoreUpdateAvailable
        if (messagingSegmentKeys.includes(MessageNotifyStoreUpdateAvailable)) {
            const entries = await durableStore.getEntries([MessageNotifyStoreUpdateAvailable], MessagingDurableSegment);
            if (entries !== undefined) {
                const notifyEntry = entries[MessageNotifyStoreUpdateAvailable];
                if (notifyEntry !== undefined) {
                    const keys = notifyEntry.data;
                    if (keys.length > 0) {
                        environment.notifyStoreUpdateAvailable(keys);
                    }
                }
            }
        }
    });
    const dispose = function () {
        validateNotDisposed();
        disposed = true;
        return unsubscribe();
    };
    const storePublish = function (key, data) {
        validateNotDisposed();
        if (stagingStore === null) {
            stagingStore = buildIngestStagingStore(environment);
        }
        stagingStore.publish(key, data);
        // remove record from main luvio L1 cache while we are on the synchronous path
        // because we do not want some other code attempting to use the
        // in-memory values before the durable store onChanged handler
        // calls back and revives the values to in-memory
        environment.storeDealloc(key);
    };
    const publishStoreMetadata = function (recordId, storeMetadata) {
        validateNotDisposed();
        if (stagingStore === null) {
            stagingStore = buildIngestStagingStore(environment);
        }
        stagingStore.publishMetadata(recordId, storeMetadata);
    };
    const storeIngest = function (key, ingest, response, luvio) {
        validateNotDisposed();
        // we don't ingest to the luvio L1 store from network directly, we ingest to
        // L2 and let DurableStore on change event revive keys into luvio L1 store
        if (stagingStore === null) {
            stagingStore = buildIngestStagingStore(environment);
        }
        environment.storeIngest(key, ingest, response, luvio, stagingStore);
    };
    const storeIngestError = function (key, errorSnapshot, storeMetadataParams, _storeOverride) {
        validateNotDisposed();
        if (stagingStore === null) {
            stagingStore = buildIngestStagingStore(environment);
        }
        environment.storeIngestError(key, errorSnapshot, storeMetadataParams, stagingStore);
    };
    const storeBroadcast = function (_rebuildSnapshot, _snapshotDataAvailable) {
        validateNotDisposed();
        // publishing to L2 is essentially "broadcasting" because the onChanged
        // handler will fire which will revive records to the main L1 store and
        // call the base storeBroadcast
        return publishChangesToDurableStore();
    };
    const publishChangesToDurableStore = function (additionalDurableStoreOperations) {
        validateNotDisposed();
        if (stagingStore === null) {
            return Promise.resolve();
        }
        const promise = flushInMemoryStoreValuesToDurableStore(stagingStore, durableStore, durableStoreErrorHandler, new Map(pendingStoreRedirects), additionalDurableStoreOperations, enableDurableMetadataRefresh);
        pendingStoreRedirects.clear();
        stagingStore = null;
        return promise;
    };
    const storeLookup = function (sel, createSnapshot, refresh, ttlStrategy) {
        validateNotDisposed();
        // if this lookup is right after an ingest or during a revive there will be a staging store
        if (stagingStore !== null) {
            const reader = new Reader(stagingStore, sel.variables, refresh, undefined, ttlStrategy);
            return reader.read(sel);
        }
        // otherwise this is from buildCachedSnapshot and we should use the luvio
        // L1 store
        return environment.storeLookup(sel, createSnapshot, refresh, ttlStrategy);
    };
    const storeEvict = function (key) {
        validateNotDisposed();
        if (stagingStore === null) {
            stagingStore = buildIngestStagingStore(environment);
        }
        stagingStore.evict(key);
    };
    const getNode = function (key) {
        validateNotDisposed();
        if (stagingStore !== null) {
            return environment.getNode(key, stagingStore);
        }
        return environment.getNode(key);
    };
    const wrapNormalizedGraphNode = function (normalized, key) {
        validateNotDisposed();
        if (stagingStore !== null) {
            return environment.wrapNormalizedGraphNode(normalized, key, stagingStore);
        }
        return environment.wrapNormalizedGraphNode(normalized, key);
    };
    const rebuildSnapshot = function (snapshot, onRebuild) {
        validateNotDisposed();
        // try rebuilding from memory
        environment.rebuildSnapshot(snapshot, (rebuilt) => {
            // only try reviving from durable store if snapshot is unfulfilled
            if (rebuilt.state !== 'Unfulfilled') {
                onRebuild(rebuilt);
                return;
            }
            // Do an L2 revive and emit to subscriber using the callback.
            reviveSnapshotWrapper(rebuilt, () => {
                // reviveSnapshot will revive into L1, and since "records" is a reference
                // (and not a copy) to the L1 records we can use it for rebuild
                let rebuiltSnap;
                environment.rebuildSnapshot(snapshot, (rebuilt) => {
                    rebuiltSnap = rebuilt;
                });
                return rebuiltSnap;
            }).then((result) => {
                onRebuild(result.snapshot);
            });
        });
    };
    const withContext = function (adapter, options) {
        validateNotDisposed();
        const { contextId, contextVersion, onContextLoaded } = options;
        let context = undefined;
        let contextKey = `${contextId}`;
        // if a context version is supplied, key with the version encoded
        if (contextVersion !== undefined) {
            contextKey += `::${contextVersion}`;
        }
        contextKey += ADAPTER_CONTEXT_ID_SUFFIX;
        const contextAsPromise = reviveOrCreateContext(contextKey, durableStore, durableStoreErrorHandler, contextStores, pendingContextStoreKeys, onContextLoaded);
        return (config, requestContext) => {
            if (context === undefined) {
                return contextAsPromise.then((revivedContext) => {
                    context = revivedContext;
                    return adapter(config, context, requestContext); // TODO - remove as any cast after https://github.com/salesforce-experience-platform-emu/luvio/pull/230
                });
            }
            return adapter(config, context, requestContext);
        };
    };
    const storeRedirect = function (existingKey, canonicalKey) {
        validateNotDisposed();
        pendingStoreRedirects.set(existingKey, canonicalKey);
        // call redirect on staging store so "old" keys are removed from L2 on
        // the next publishChangesToDurableStore.  NOTE: we don't need to call
        // redirect on the base environment store because staging store and base
        // L1 store share the same redirect and reverseRedirectKeys
        if (stagingStore === null) {
            stagingStore = buildIngestStagingStore(environment);
        }
        stagingStore.redirect(existingKey, canonicalKey);
    };
    const storeSetTTLOverride = function (namespace, representationName, ttl) {
        validateNotDisposed();
        return Promise.all([
            environment.storeSetTTLOverride(namespace, representationName, ttl),
            durableTTLStore.setDurableTTLOverride(namespace, representationName, ttl),
        ]).then();
    };
    const storeSetDefaultTTLOverride = function (ttl) {
        validateNotDisposed();
        return Promise.all([
            environment.storeSetDefaultTTLOverride(ttl),
            durableTTLStore.setDefaultDurableTTLOverrides(ttl),
        ]).then();
    };
    const getDurableTTLOverrides = function () {
        validateNotDisposed();
        return durableTTLStore.getDurableTTLOverrides();
    };
    const dispatchResourceRequest = async function (request, context, eventObservers) {
        validateNotDisposed();
        // non-GET adapters call dispatchResourceRequest before any other luvio
        // function so this is our chance to ensure we're initialized
        if (initializationPromise !== undefined) {
            await initializationPromise;
        }
        return environment.dispatchResourceRequest(request, context, eventObservers);
    };
    // NOTE: we can't use "async" keyword on this function because that would
    // force it to always be an async response.  The signature is a union
    // of sync/async so no "awaiting" in this function, just promise-chaining
    const applyCachePolicy = function (luvio, adapterRequestContext, buildSnapshotContext, buildCachedSnapshot, buildNetworkSnapshot) {
        validateNotDisposed();
        const wrappedCacheLookup = (injectedBuildSnapshotContext, injectedStoreLookup) => {
            const snapshot = buildCachedSnapshot(injectedBuildSnapshotContext, injectedStoreLookup, luvio);
            // if the adapter attempted to do an L1 lookup and it was unfulfilled
            // then we can attempt an L2 lookup
            if (isUnfulfilledSnapshot(snapshot)) {
                const start = Date.now();
                emitDurableEnvironmentAdapterEvent({ type: 'l2-revive-start' }, adapterRequestContext.eventObservers);
                const revivedSnapshot = reviveSnapshotWrapper(snapshot, () => injectedStoreLookup(snapshot.select, snapshot.refresh)).then((result) => {
                    emitDurableEnvironmentAdapterEvent({
                        type: 'l2-revive-end',
                        snapshot: result.snapshot,
                        duration: Date.now() - start,
                        l2Trips: result.metrics.l2Trips,
                    }, adapterRequestContext.eventObservers);
                    return result.snapshot;
                });
                return revivedSnapshot;
            }
            // otherwise just return what buildCachedSnapshot gave us
            return snapshot;
        };
        const wrappedApplyCachePolicy = () => {
            return environment.applyCachePolicy(luvio, adapterRequestContext, buildSnapshotContext, wrappedCacheLookup, buildNetworkSnapshot);
        };
        // GET adapters call applyCachePolicy before any other luvio
        // function so this is our chance to ensure we're initialized
        return initializationPromise !== undefined
            ? initializationPromise.then(wrappedApplyCachePolicy)
            : wrappedApplyCachePolicy();
    };
    const getIngestStagingStoreRecords = function () {
        validateNotDisposed();
        if (stagingStore !== null) {
            return stagingStore.fallbackStringKeyInMemoryStore.records;
        }
        return {};
    };
    const getIngestStagingStoreMetadata = function () {
        validateNotDisposed();
        if (stagingStore !== null) {
            return stagingStore.fallbackStringKeyInMemoryStore.metadata;
        }
        return {};
    };
    const getIngestStagingStore = function () {
        validateNotDisposed();
        return stagingStore === null || stagingStore === void 0 ? void 0 : stagingStore.fallbackStringKeyInMemoryStore;
    };
    const handleSuccessResponse = async function (ingestAndBroadcastFunc, getResponseCacheKeysFunc) {
        validateNotDisposed();
        const cacheKeyMap = getResponseCacheKeysFunc();
        const cacheKeyMapKeys = cacheKeyMap.keysAsArray();
        const keysToRevive = new StoreKeySet();
        for (const cacheKeyMapKey of cacheKeyMapKeys) {
            const cacheKey = cacheKeyMap.get(cacheKeyMapKey);
            if (cacheKey.mergeable === true) {
                const canonical = environment.storeGetCanonicalKey(cacheKeyMapKey);
                keysToRevive.add(canonical);
            }
        }
        let snapshotFromMemoryIngest = undefined;
        // To-do: Once these are structured keys, will need to support them throughout durable logic W-12356727
        const keysToReviveAsArray = Array.from(keysToRevive.keysAsStrings());
        if (keysToReviveAsArray.length > 0) {
            // if we need to do an L2 read then L2 write then we need to synchronize
            // our read/merge/ingest/write Promise based on the keys so we don't
            // stomp over any data
            const readWritePromise = (async () => {
                const pendingPromises = [];
                for (const key of keysToReviveAsArray) {
                    const pendingPromise = mergeKeysPromiseMap.get(key);
                    if (pendingPromise !== undefined) {
                        // IMPORTANT: while on the synchronous code path we get a
                        // handle to pendingPromise and push it onto the array.
                        // This is important because later in this synchronous code
                        // path we will upsert readWritePromise into the
                        // mergeKeysPromiseMap (essentially overwriting pendingPromise
                        // in the map).
                        pendingPromises.push(pendingPromise);
                    }
                }
                await Promise.all(pendingPromises);
                const entries = await durableStore.getEntries(keysToReviveAsArray, DefaultDurableSegment);
                stagingStore = buildIngestStagingStore(environment);
                publishDurableStoreEntries(entries, (key, record) => {
                    if (typeof key === 'string') {
                        stagingStore.fallbackStringKeyInMemoryStore.records[key] = record;
                    }
                    else {
                        stagingStore.recordsMap.set(key, record);
                    }
                }, (key, metadata) => {
                    if (typeof key === 'string') {
                        stagingStore.fallbackStringKeyInMemoryStore.metadata[key] = metadata;
                    }
                    else {
                        stagingStore.metadataMap.set(key, metadata);
                    }
                });
                snapshotFromMemoryIngest = await ingestAndBroadcastFunc();
            })();
            for (const key of keysToReviveAsArray) {
                // we are overwriting the previous promise at this key, but that
                // is ok because we got a handle to it earlier (see the IMPORTANT
                // comment about 35 lines up)
                mergeKeysPromiseMap.set(key, readWritePromise);
            }
            try {
                await readWritePromise;
            }
            finally {
                for (const key of keysToReviveAsArray) {
                    const pendingPromise = mergeKeysPromiseMap.get(key);
                    // cleanup the entry from the map if this is the last promise
                    // for that key
                    if (pendingPromise === readWritePromise) {
                        mergeKeysPromiseMap.delete(key);
                    }
                }
            }
        }
        else {
            // we aren't doing any merging so we don't have to synchronize, the
            // underlying DurableStore implementation takes care of R/W sync
            // so all we have to do is ingest then write to L2
            stagingStore = buildIngestStagingStore(environment);
            snapshotFromMemoryIngest = await ingestAndBroadcastFunc();
        }
        if (snapshotFromMemoryIngest === undefined) {
            return undefined;
        }
        if (snapshotFromMemoryIngest.state !== 'Unfulfilled') {
            return snapshotFromMemoryIngest;
        }
        // if snapshot from staging store lookup is unfulfilled then do an L2 lookup
        const { select, refresh } = snapshotFromMemoryIngest;
        const result = await reviveSnapshotWrapper(snapshotFromMemoryIngest, () => environment.storeLookup(select, environment.createSnapshot, refresh));
        return result.snapshot;
    };
    const handleErrorResponse = async function (ingestAndBroadcastFunc) {
        validateNotDisposed();
        stagingStore = buildIngestStagingStore(environment);
        return ingestAndBroadcastFunc();
    };
    const getNotifyChangeStoreEntries = function (keys) {
        validateNotDisposed();
        return durableStore
            .getEntries(keys.map(serializeStructuredKey), DefaultDurableSegment)
            .then((durableRecords) => {
            const entries = [];
            publishDurableStoreEntries(durableRecords, (key, record) => {
                entries.push({
                    key,
                    record: record,
                });
            }, () => { });
            return entries;
        });
    };
    // flag entries specified by keys in durable store as expired.
    // send a notifyStoreUpdateAvailable message through durable store set entries to
    // indirectly triggers all js environments to refresh snapshots if overlapping with keys.
    const notifyStoreUpdateAvailable = async function (keys$1) {
        validateNotDisposed();
        const entryKeys = keys$1.map(serializeStructuredKey);
        const entries = await durableStore.getEntries(entryKeys, DefaultDurableSegment);
        if (entries === undefined || keys$3(entries).length === 0) {
            return environment.notifyStoreUpdateAvailable(keys$1);
        }
        const now = Date.now();
        let needWriteBack = false;
        for (let i = 0; i < entryKeys.length; i++) {
            const key = entryKeys[i];
            const entry = entries[key];
            if (entry !== undefined) {
                const storeEntry = entry;
                if (storeEntry.metadata !== undefined) {
                    storeEntry.metadata = {
                        ...storeEntry.metadata,
                        expirationTimestamp: now,
                    };
                }
                needWriteBack = true;
            }
        }
        if (needWriteBack) {
            await durableStore.setEntries(entries, DefaultDurableSegment);
        }
        // push a notifyStoreUpdateAvailable message with entryKeys as data into messaging segment
        await durableStore.setEntries({ notifyStoreUpdateAvailable: { data: entryKeys } }, MessagingDurableSegment);
        return Promise.resolve(undefined);
    };
    const reviveSnapshotWrapper = function (unavailableSnapshot, buildL1Snapshot) {
        let revivingStore = undefined;
        if (useRevivingStore) {
            // NOTE: `store` is private, there doesn't seem to be a better,
            // cleaner way of accessing it from a derived environment.
            let baseStore = environment.store;
            // If we're rebuilding during an ingest, the existing staging store should be the base store.
            if (stagingStore) {
                baseStore = stagingStore;
            }
            let revivingStore = buildRevivingStagingStore(baseStore);
            revivingStores.add(revivingStore);
        }
        return reviveSnapshot(environment, durableStore, unavailableSnapshot, durableStoreErrorHandler, () => {
            const tempStore = stagingStore;
            const result = buildL1Snapshot();
            stagingStore = tempStore;
            return result;
        }, revivingStore).finally(() => {
        });
    };
    // set the default cache policy of the base environment
    environment.setDefaultCachePolicy({
        type: 'stale-while-revalidate',
        staleDurationSeconds: Number.MAX_SAFE_INTEGER,
    });
    return create$3(environment, {
        publishStoreMetadata: { value: publishStoreMetadata },
        storeIngest: { value: storeIngest },
        storeIngestError: { value: storeIngestError },
        storeBroadcast: { value: storeBroadcast },
        storeLookup: { value: storeLookup },
        storeEvict: { value: storeEvict },
        wrapNormalizedGraphNode: { value: wrapNormalizedGraphNode },
        getNode: { value: getNode },
        rebuildSnapshot: { value: rebuildSnapshot },
        withContext: { value: withContext },
        storeSetTTLOverride: { value: storeSetTTLOverride },
        storeSetDefaultTTLOverride: { value: storeSetDefaultTTLOverride },
        storePublish: { value: storePublish },
        storeRedirect: { value: storeRedirect },
        dispose: { value: dispose },
        publishChangesToDurableStore: { value: publishChangesToDurableStore },
        getDurableTTLOverrides: { value: getDurableTTLOverrides },
        dispatchResourceRequest: { value: dispatchResourceRequest },
        applyCachePolicy: { value: applyCachePolicy },
        getIngestStagingStoreRecords: { value: getIngestStagingStoreRecords },
        getIngestStagingStoreMetadata: { value: getIngestStagingStoreMetadata },
        getIngestStagingStore: { value: getIngestStagingStore },
        handleSuccessResponse: { value: handleSuccessResponse },
        handleErrorResponse: { value: handleErrorResponse },
        getNotifyChangeStoreEntries: { value: getNotifyChangeStoreEntries },
        notifyStoreUpdateAvailable: { value: notifyStoreUpdateAvailable },
    });
}

const { keys: keys$2, create: create$2, assign: assign$2, entries: entries$1, values: values$2 } = Object;
const { stringify: stringify$1, parse: parse$1 } = JSON;

function selectColumnsFromTableWhereKeyIn(columnNames, table, keyColumnName, whereIn) {
    const paramList = whereIn.map(() => '?').join(',');
    return `SELECT ${columnNames.join(',')} FROM ${table} WHERE ${keyColumnName} IN (${paramList})`;
}
function selectColumnsFromTableWhereKeyInNamespaced(columnNames, table, keyColumnName, whereIn, namespaceColumnName) {
    const paramList = whereIn.map(() => '?').join(',');
    return `SELECT ${columnNames.join(',')} FROM ${table} WHERE ${namespaceColumnName} = ? AND ${keyColumnName} IN (${paramList})`;
}

// These const values must be in sync with the latest
// @salesforce/nimbus-plugin-lds/sql schema file
const TABLE_NAME$1 = 'lds_data';
const COLUMN_NAME_KEY$2 = 'key';
const COLUMN_NAME_DATA$2 = 'data';
const COLUMN_NAME_METADATA$1 = 'metadata';
class LdsDataTable {
    constructor(plugin) {
        this.tableName = TABLE_NAME$1;
        this.columnNames = [COLUMN_NAME_KEY$2, COLUMN_NAME_DATA$2, COLUMN_NAME_METADATA$1];
        this.conflictColumnNames = [COLUMN_NAME_KEY$2];
        this.getAllQuery = `SELECT ${this.columnNames.join(',')} FROM ${this.tableName}`;
        this.plugin = plugin;
    }
    getByKeys(keys) {
        const getQuery = selectColumnsFromTableWhereKeyIn(this.columnNames, this.tableName, COLUMN_NAME_KEY$2, keys);
        return new Promise((resolve, reject) => {
            this.plugin.query(getQuery, keys, (x) => {
                resolve(this.mapToDurableEntries(x));
            }, reject);
        });
    }
    getMetadataByKeys(keys) {
        const query = selectColumnsFromTableWhereKeyIn([COLUMN_NAME_KEY$2, COLUMN_NAME_METADATA$1], this.tableName, COLUMN_NAME_KEY$2, keys);
        return new Promise((resolve, reject) => {
            this.plugin.query(query, keys, (results) => {
                resolve(results.rows.reduce((entries, row) => {
                    const [key, stringifiedMetadata] = row;
                    if (stringifiedMetadata !== undefined) {
                        entries[key] = {
                            metadata: parse$1(stringifiedMetadata),
                        };
                    }
                    return entries;
                }, {}));
            }, reject);
        });
    }
    getAll() {
        return new Promise((resolve, reject) => {
            this.plugin.query(this.getAllQuery, [], (x) => {
                resolve(this.mapToDurableEntries(x));
            }, reject);
        });
    }
    entriesToUpsertOperations(entries, segment) {
        return {
            type: 'upsert',
            table: this.tableName,
            keyColumn: COLUMN_NAME_KEY$2,
            context: {
                segment,
            },
            conflictColumns: this.conflictColumnNames,
            columns: this.columnNames,
            rows: keys$2(entries).reduce((rows, key) => {
                const entry = entries[key];
                const { data, metadata } = entry;
                const row = [key, stringify$1(data), metadata ? stringify$1(metadata) : null];
                rows.push(row);
                return rows;
            }, []),
        };
    }
    metadataToUpdateOperations(entries, segment) {
        return {
            type: 'update',
            table: this.tableName,
            keyColumn: COLUMN_NAME_KEY$2,
            context: {
                segment,
                type: 'setMetadata',
            },
            columns: [COLUMN_NAME_METADATA$1],
            values: keys$2(entries).reduce((values, key) => {
                const { metadata } = entries[key];
                const row = [metadata ? stringify$1(metadata) : null];
                values[key] = row;
                return values;
            }, {}),
        };
    }
    mapToDurableEntries(sqliteResult) {
        return sqliteResult.rows.reduce((entries, row) => {
            const [key, stringifiedData, stringifiedMetadata] = row;
            const durableStoreEntry = {
                data: parse$1(stringifiedData),
            };
            if (stringifiedMetadata !== null) {
                durableStoreEntry.metadata = parse$1(stringifiedMetadata);
            }
            entries[key] = durableStoreEntry;
            return entries;
        }, {});
    }
}

// These const values must be in sync with the latest
// @salesforce/nimbus-plugin-lds/sql schema file
const TABLE_NAME = 'lds_internal';
const COLUMN_NAME_KEY$1 = 'key';
const COLUMN_NAME_DATA$1 = 'data';
const COLUMN_NAME_METADATA = 'metadata';
const COLUMN_NAME_NAMESPACE = 'namespace';
class LdsInternalDataTable {
    constructor(plugin) {
        this.tableName = TABLE_NAME;
        this.columnNames = [
            COLUMN_NAME_KEY$1,
            COLUMN_NAME_DATA$1,
            COLUMN_NAME_METADATA,
            COLUMN_NAME_NAMESPACE,
        ];
        this.conflictColumnNames = [COLUMN_NAME_KEY$1, COLUMN_NAME_NAMESPACE];
        this.getAllQuery = `SELECT ${this.columnNames.join(',')} FROM ${this.tableName} WHERE ${COLUMN_NAME_NAMESPACE} = ?`;
        this.plugin = plugin;
    }
    getByKeys(keys, namespace) {
        if (namespace === undefined) {
            throw Error('LdsInternalDataTable requires namespace');
        }
        const getQuery = selectColumnsFromTableWhereKeyInNamespaced(this.columnNames, this.tableName, COLUMN_NAME_KEY$1, keys, COLUMN_NAME_NAMESPACE);
        return new Promise((resolve, reject) => {
            this.plugin.query(getQuery, [namespace].concat(keys), (x) => {
                resolve(this.mapToDurableEntries(x));
            }, reject);
        });
    }
    getMetadataByKeys(keys, namespace) {
        if (namespace === undefined) {
            throw Error('LdsInternalDataTable requires namespace');
        }
        const query = selectColumnsFromTableWhereKeyInNamespaced([COLUMN_NAME_KEY$1, COLUMN_NAME_METADATA], this.tableName, COLUMN_NAME_KEY$1, keys, COLUMN_NAME_NAMESPACE);
        return new Promise((resolve, reject) => {
            this.plugin.query(query, [namespace].concat(keys), (results) => {
                resolve(results.rows.reduce((entries, row) => {
                    const [key, stringifiedMetadata] = row;
                    if (stringifiedMetadata !== undefined) {
                        entries[key] = {
                            metadata: parse$1(stringifiedMetadata),
                        };
                    }
                    return entries;
                }, {}));
            }, reject);
        });
    }
    getAll(namespace) {
        return new Promise((resolve, reject) => {
            this.plugin.query(this.getAllQuery, [namespace], (x) => {
                resolve(this.mapToDurableEntries(x));
            }, reject);
        });
    }
    entriesToUpsertOperations(entries, segment) {
        return {
            type: 'upsert',
            table: this.tableName,
            keyColumn: COLUMN_NAME_KEY$1,
            context: {
                segment,
            },
            conflictColumns: this.conflictColumnNames,
            columns: this.columnNames,
            rows: keys$2(entries).reduce((rows, key) => {
                const entry = entries[key];
                const { data, metadata } = entry;
                const row = [key, stringify$1(data)];
                if (metadata) {
                    row.push(stringify$1(metadata));
                }
                else {
                    row.push(null);
                }
                row.push(segment);
                rows.push(row);
                return rows;
            }, []),
        };
    }
    metadataToUpdateOperations(entries, segment) {
        return {
            type: 'update',
            table: this.tableName,
            keyColumn: COLUMN_NAME_KEY$1,
            context: {
                segment,
                type: 'setMetadata',
            },
            columns: [COLUMN_NAME_METADATA],
            values: keys$2(entries).reduce((values, key) => {
                const { metadata } = entries[key];
                const row = [metadata ? stringify$1(metadata) : null];
                values[key] = row;
                return values;
            }, {}),
        };
    }
    metadataToUpdateSQLQueries(entries, segment) {
        return keys$2(entries).reduce((accu, key) => {
            const { metadata } = entries[key];
            if (metadata !== undefined) {
                accu.push({
                    sql: `UPDATE ${this.tableName} SET ${COLUMN_NAME_METADATA} = ? WHERE (${COLUMN_NAME_KEY$1} IS ? AND ${COLUMN_NAME_NAMESPACE} IS ?)`,
                    params: [stringify$1(metadata), key, segment],
                    change: {
                        ids: [key],
                        segment,
                        type: 'setMetadata',
                        isExternalChange: false,
                    },
                });
            }
            return accu;
        }, []);
    }
    mapToDurableEntries(sqliteResult) {
        return sqliteResult.rows.reduce((entries, row) => {
            const [key, stringifiedData, stringifiedMetadata] = row;
            const durableStoreEntry = {
                data: parse$1(stringifiedData),
            };
            if (stringifiedMetadata !== null) {
                durableStoreEntry.metadata = parse$1(stringifiedMetadata);
            }
            entries[key] = durableStoreEntry;
            return entries;
        }, {});
    }
}

class NimbusSqliteStore {
    constructor(plugin, additionalTableMap = {}) {
        this.plugin = plugin;
        this.internalDataTable = new LdsInternalDataTable(plugin);
        this.dataTableMap = {
            ...additionalTableMap,
            [DefaultDurableSegment]: new LdsDataTable(plugin),
        };
        this.supportsBatchUpdates = plugin.supportsBatchUpdates !== undefined;
    }
    isEvalSupported() {
        return true;
    }
    isBatchUpdateSupported() {
        return this.supportsBatchUpdates;
    }
    query(sql, params) {
        return new Promise((resolve, reject) => {
            this.plugin.query(sql, params, (result) => {
                resolve(result);
            }, (error) => {
                reject(error);
            });
        });
    }
    batchQuery(queries) {
        const promises = queries.map((q) => this.query(q.sql, q.params));
        return Promise.all(promises);
    }
    async getEntries(entryIds, segment) {
        return this.getTable(segment).getByKeys(entryIds, segment);
    }
    async getMetadata(entryIds, segment) {
        return this.getTable(segment).getMetadataByKeys(entryIds, segment);
    }
    getAllEntries(segment) {
        return this.getTable(segment).getAll(segment);
    }
    setEntries(entries, segment) {
        if (keys$2(entries).length === 0) {
            return Promise.resolve();
        }
        const table = this.getTable(segment);
        const upsertOperation = table.entriesToUpsertOperations(entries, segment);
        return this.batchOperationAsPromise([upsertOperation]);
    }
    setMetadata(entries, segment) {
        if (keys$2(entries).length === 0) {
            return Promise.resolve();
        }
        const table = this.getTable(segment);
        let operation;
        if (this.supportsBatchUpdates) {
            operation = table.metadataToUpdateOperations(entries, segment);
        }
        else {
            operation = table.entriesToUpsertOperations(entries, segment);
            // manually set the context type on the upsert so notifications do not notify rebuilds without
            // plugin updates
            operation.context.type = 'setMetadata';
        }
        return this.batchOperationAsPromise([operation]);
    }
    batchOperations(operations) {
        const sqliteOperations = operations.reduce((acc, cur) => {
            if (cur.type === 'setEntries') {
                if (keys$2(cur.entries).length > 0) {
                    const table = this.getTable(cur.segment);
                    acc.push(table.entriesToUpsertOperations(cur.entries, cur.segment));
                }
            }
            else if (cur.type === 'setMetadata') {
                if (keys$2(cur.entries).length > 0) {
                    const table = this.getTable(cur.segment);
                    if (this.supportsBatchUpdates) {
                        acc.push(table.metadataToUpdateOperations(cur.entries, cur.segment));
                    }
                    else {
                        const upsert = table.entriesToUpsertOperations(cur.entries, cur.segment);
                        // manually set the context type on the upsert so notifications do not notify rebuilds without
                        // plugin updates
                        upsert.context.type = 'setMetadata';
                        acc.push(upsert);
                    }
                }
            }
            else {
                if (cur.ids.length > 0) {
                    acc.push(this.idsToDeleteOperation(cur.ids, cur.segment));
                }
            }
            return acc;
        }, []);
        return sqliteOperations.length === 0
            ? Promise.resolve()
            : this.batchOperationAsPromise(sqliteOperations);
    }
    evictEntries(entryIds, segment) {
        return entryIds.length === 0
            ? Promise.resolve()
            : this.batchOperationAsPromise([this.idsToDeleteOperation(entryIds, segment)]);
    }
    registerOnChangedListener(listener) {
        let unsubscribeId = undefined;
        this.plugin
            .registerOnChangedListener(async (changes) => {
            const durableChanges = changes.map((c) => {
                let type = c.type === 'upsert' ? 'setEntries' : 'evictEntries';
                // if our context contains a type then set that as our main level type
                // allows us in the future of updates to specify the segment change happening
                // example being update call on metadata only or updating data
                if ((c.type === 'update' || c.type === 'upsert') &&
                    c.context.type !== undefined) {
                    type = c.context.type;
                }
                return {
                    type,
                    ids: c.keys,
                    isExternalChange: false,
                    segment: c.context.segment,
                };
            });
            await listener(durableChanges);
        })
            .then((unsub) => {
            unsubscribeId = unsub;
        });
        return () => {
            if (unsubscribeId) {
                return this.plugin.unsubscribeOnChangedListener(unsubscribeId);
            }
            return Promise.resolve();
        };
    }
    getTable(segment) {
        var _a;
        return (_a = this.dataTableMap[segment]) !== null && _a !== void 0 ? _a : this.internalDataTable;
    }
    idsToDeleteOperation(entryIds, segment) {
        const table = this.getTable(segment);
        return {
            type: 'delete',
            table: table.tableName,
            keyColumn: 'key',
            context: {
                segment,
            },
            ids: entryIds,
        };
    }
    batchOperationAsPromise(sqliteOperations) {
        return new Promise((resolve, reject) => {
            this.plugin.batchOperations(sqliteOperations, (error) => {
                if (error && error !== null) {
                    reject(error);
                }
                else {
                    resolve();
                }
            });
        });
    }
}

/**
 * Copyright (c) 2022, Salesforce, Inc.,
 * All rights reserved.
 * For full license text, see the LICENSE.txt file
 */


function isStoreKeyRecordId(key) {
    return key.indexOf(RECORD_ID_PREFIX) > -1 && key.indexOf(RECORD_FIELDS_KEY_JUNCTION) === -1;
}
function buildRecordFieldStoreKey(recordKey, fieldName) {
    return `${recordKey}${RECORD_FIELDS_KEY_JUNCTION}${fieldName}`;
}

function isStoreRecordError(storeRecord) {
    return storeRecord.__type === 'error';
}
function isEntryDurableRecordRepresentation(entry, key) {
    // Either a DurableRecordRepresentation or StoreRecordError can live at a record key
    return ((isStoreKeyRecordId(key) || isStoreKeyRecordViewEntity(key)) &&
        entry.data.__type === undefined);
}

/**
 * Copyright (c) 2022, Salesforce, Inc.,
 * All rights reserved.
 * For full license text, see the LICENSE.txt file
 */


var DraftActionStatus;
(function (DraftActionStatus) {
    DraftActionStatus["Pending"] = "pending";
    DraftActionStatus["Uploading"] = "uploading";
    DraftActionStatus["Error"] = "error";
    DraftActionStatus["Completed"] = "completed";
})(DraftActionStatus || (DraftActionStatus = {}));
var ProcessActionResult;
(function (ProcessActionResult) {
    // non-2xx network error, requires user intervention
    ProcessActionResult["ACTION_ERRORED"] = "ERROR";
    // upload succeeded
    ProcessActionResult["ACTION_SUCCEEDED"] = "SUCCESS";
    // queue is empty
    ProcessActionResult["NO_ACTION_TO_PROCESS"] = "NO_ACTION_TO_PROCESS";
    // network request is in flight
    ProcessActionResult["ACTION_ALREADY_PROCESSING"] = "ACTION_ALREADY_PROCESSING";
    // network call failed (offline)
    ProcessActionResult["NETWORK_ERROR"] = "NETWORK_ERROR";
    // queue is blocked on an error that requires user intervention
    ProcessActionResult["BLOCKED_ON_ERROR"] = "BLOCKED_ON_ERROR";
    //waiting for user to execute custom action
    ProcessActionResult["CUSTOM_ACTION_WAITING"] = "CUSTOM_ACTION_WAITING";
})(ProcessActionResult || (ProcessActionResult = {}));
var DraftQueueState;
(function (DraftQueueState) {
    /** Currently processing an item in the queue or queue is empty and waiting to process the next item. */
    DraftQueueState["Started"] = "started";
    /**
     * The queue is stopped and will not attempt to upload any drafts until startDraftQueue() is called.
     * This is the initial state when the DraftQueue gets instantiated.
     */
    DraftQueueState["Stopped"] = "stopped";
    /**
     * The queue is stopped due to a blocking error from the last upload attempt.
     * The queue will not run again until startDraftQueue() is called.
     */
    DraftQueueState["Error"] = "error";
    /**
     * There was a network error and the queue will attempt to upload again shortly.
     * To attempt to force an upload now call startDraftQueue().
     */
    DraftQueueState["Waiting"] = "waiting";
})(DraftQueueState || (DraftQueueState = {}));
var DraftQueueEventType;
(function (DraftQueueEventType) {
    /**
     * Triggered after an action had been added to the queue
     */
    DraftQueueEventType["ActionAdded"] = "added";
    /**
     * Triggered when starting to upload and process an action
     */
    DraftQueueEventType["ActionUploading"] = "uploading";
    /**
     * Triggered once an action failed
     */
    DraftQueueEventType["ActionFailed"] = "failed";
    /**
     * Triggered after an action has been deleted from the queue
     */
    DraftQueueEventType["ActionDeleted"] = "deleted";
    /**
     * Triggered after an action has been completed and after it has been removed from the queue
     */
    DraftQueueEventType["ActionCompleted"] = "completed";
    /**
     * Triggered after an action has been updated by the updateAction API
     */
    DraftQueueEventType["ActionUpdated"] = "updated";
    /**
     * Triggered after the Draft Queue state changes
     */
    DraftQueueEventType["QueueStateChanged"] = "state";
})(DraftQueueEventType || (DraftQueueEventType = {}));
var QueueOperationType;
(function (QueueOperationType) {
    QueueOperationType["Add"] = "add";
    QueueOperationType["Delete"] = "delete";
    QueueOperationType["Update"] = "update";
})(QueueOperationType || (QueueOperationType = {}));

var CustomActionResultType;
(function (CustomActionResultType) {
    CustomActionResultType["SUCCESS"] = "SUCCESS";
    CustomActionResultType["FAILURE"] = "FAILURE";
})(CustomActionResultType || (CustomActionResultType = {}));
var CustomActionErrorType;
(function (CustomActionErrorType) {
    CustomActionErrorType["NETWORK_ERROR"] = "NETWORK_ERROR";
    CustomActionErrorType["CLIENT_ERROR"] = "CLIENT_ERROR";
})(CustomActionErrorType || (CustomActionErrorType = {}));

/**
 * Denotes what kind of operation a DraftQueueItem represents.
 */
var DraftActionOperationType;
(function (DraftActionOperationType) {
    DraftActionOperationType["Create"] = "create";
    DraftActionOperationType["Update"] = "update";
    DraftActionOperationType["Delete"] = "delete";
    DraftActionOperationType["Custom"] = "custom";
})(DraftActionOperationType || (DraftActionOperationType = {}));
var DraftQueueOperationType;
(function (DraftQueueOperationType) {
    DraftQueueOperationType["ItemAdded"] = "added";
    DraftQueueOperationType["ItemUploading"] = "uploading";
    DraftQueueOperationType["ItemDeleted"] = "deleted";
    DraftQueueOperationType["ItemCompleted"] = "completed";
    DraftQueueOperationType["ItemFailed"] = "failed";
    DraftQueueOperationType["ItemUpdated"] = "updated";
    DraftQueueOperationType["QueueStarted"] = "started";
    DraftQueueOperationType["QueueStopped"] = "stopped";
})(DraftQueueOperationType || (DraftQueueOperationType = {}));

function serializeFieldArguments(argumentNodes, variables) {
    const mutableArgumentNodes = Object.assign([], argumentNodes);
    return `args__(${mutableArgumentNodes
        .sort((a, b) => {
        const aName = a.name.value.toUpperCase();
        const bName = b.name.value.toUpperCase();
        return aName < bName ? -1 : aName > bName ? 1 : 0;
    })
        .map((node) => serializeArgNode(node, variables))
        .join('::')})`;
}
function serializeArgNode(argumentNode, variables) {
    const argName = argumentNode.name.value;
    return `${argName}:${serializeValueNode(argumentNode.value, variables)}`;
}
function serializeValueNode(valueNode, variables) {
    switch (valueNode.kind) {
        case 'BooleanValue':
            return valueNode.value + '';
        case 'IntValue':
        case 'FloatValue':
        case 'EnumValue':
        case 'StringValue':
            return valueNode.value;
        case 'ListValue': {
            const mutableValueNodeList = Object.assign([], valueNode.values);
            return mutableValueNodeList
                .sort((a, b) => {
                const aVal = serializeValueNode(a, variables).toUpperCase();
                const bVal = serializeValueNode(b, variables).toUpperCase();
                return aVal < bVal ? -1 : aVal > bVal ? 1 : 0;
            })
                .map((val, i) => `${serializeValueNode(val, variables)}[${i}]`)
                .join(',');
        }
        case 'Variable': {
            const variableValue = variables[valueNode.name.value];
            return typeof variableValue === 'string'
                ? variableValue
                : JSON.stringify(variableValue);
        }
        case 'NullValue':
            return 'null';
        case 'ObjectValue': {
            const mutableFieldNodeList = Object.assign([], valueNode.fields);
            return mutableFieldNodeList
                .sort((a, b) => {
                const aName = a.name.value.toUpperCase();
                const bName = b.name.value.toUpperCase();
                return aName < bName ? -1 : aName > bName ? 1 : 0;
            })
                .map((field) => field.name.value + ':' + serializeValueNode(field.value, variables))
                .join(',');
        }
    }
}

function serializeOperationNode(operationNode, variables, fragmentMap) {
    return `${serializeSelectionSet(operationNode.selectionSet, variables, fragmentMap)}`;
}
function serializeSelectionSet(selectionSetNode, variables, fragmentMap) {
    return `${selectionSetNode.selections
        .map((selection) => serializeSelectionNode(selection, variables, fragmentMap))
        .join()}`;
}
/**
 *
 * @description This function takes a GraphQL SelectionNode from an AST and serializes it in a stable way, so we can
 * use it for property names and Store keys.
 * @param selectionNode
 * @param variables
 * @param fragmentMap
 * @returns string
 */
function serializeSelectionNode(selectionNode, variables, fragmentMap) {
    switch (selectionNode.kind) {
        case 'Field': {
            const hasArguments = selectionNode.arguments !== undefined && selectionNode.arguments.length > 0;
            const argumentSuffix = hasArguments
                ? `__${serializeFieldArguments(selectionNode.arguments, variables)}`
                : '';
            return `${selectionNode.name.value}${argumentSuffix}`;
        }
        case 'FragmentSpread': {
            const fragment = fragmentMap[selectionNode.name.value];
            return fragment === undefined
                ? selectionNode.name.value
                : serializeSelectionSet(fragment.selectionSet, variables, fragmentMap);
        }
        case 'InlineFragment':
            return serializeSelectionSet(selectionNode.selectionSet, variables, fragmentMap);
    }
}
function buildQueryTypeStringKey(args) {
    const { keyPrefix, schemaName, queryTypeName, operationNode, variables, fragmentMap } = args;
    return `${keyPrefix}::${schemaName}::${queryTypeName}[${serializeOperationNode(operationNode, variables, fragmentMap)}]`;
}

/**
 * Copyright (c) 2022, Salesforce, Inc.,
 * All rights reserved.
 * For full license text, see the LICENSE.txt file
 */


const { create: create$1, keys: keys$1, values: values$1, entries, assign: assign$1 } = Object;
const { stringify, parse } = JSON;
const { isArray, from } = Array;

function removeUndefined(array) {
    return array.reduce((accu, item) => {
        if (item !== undefined) {
            accu.push(item);
        }
        return accu;
    }, []);
}

var DateLiteral;
(function (DateLiteral) {
    DateLiteral["NEXT_YEAR"] = "NEXT_YEAR";
    DateLiteral["LAST_WEEK"] = "LAST_WEEK";
    DateLiteral["THIS_MONTH"] = "THIS_MONTH";
    DateLiteral["THIS_QUARTER"] = "THIS_QUARTER";
    DateLiteral["LAST_MONTH"] = "LAST_MONTH";
    DateLiteral["LAST_90_DAYS"] = "LAST_90_DAYS";
    DateLiteral["LAST_QUARTER"] = "LAST_QUARTER";
    DateLiteral["THIS_YEAR"] = "THIS_YEAR";
    DateLiteral["NEXT_QUARTER"] = "NEXT_QUARTER";
    DateLiteral["NEXT_MONTH"] = "NEXT_MONTH";
    DateLiteral["TOMORROW"] = "TOMORROW";
    DateLiteral["TODAY"] = "TODAY";
    DateLiteral["LAST_YEAR"] = "LAST_YEAR";
    DateLiteral["NEXT_90_DAYS"] = "NEXT_90_DAYS";
    DateLiteral["THIS_WEEK"] = "THIS_WEEK";
    DateLiteral["YESTERDAY"] = "YESTERDAY";
    DateLiteral["NEXT_WEEK"] = "NEXT_WEEK";
})(DateLiteral || (DateLiteral = {}));
// ranged adjusted to be between now and the end of the day,
// you cannot do a straight = comparison in SQL
const TODAY_RANGE = {
    lowerBound: `date('now')`,
    upperBound: `date('now', '+1 day', '-0.001 seconds')`,
};
// ranged adjusted to be between today +1 day and +2 days -.001 seconds to be the very end of tomorrows date
// you cannot do a straight = comparison in SQL
const TOMORROW_RANGE = {
    lowerBound: `date('now', '+1 day')`,
    upperBound: `date('now', '+2 day', '-0.001 seconds')`,
};
const YESTERDAY_RANGE = {
    lowerBound: `date('now', '-1 day')`,
    upperBound: `date('now', '-0.001 seconds')`,
};
const THIS_WEEK_RANGE = {
    lowerBound: `date('now', 'weekday 0', '-7 days')`,
    upperBound: `date('now', 'weekday 0', '-1 day')`,
};
const LAST_WEEK_RANGE = {
    lowerBound: `date('now', 'weekday 0', '-14 days')`,
    upperBound: `date('now', 'weekday 0', '-8 days')`,
};
const NEXT_WEEK_RANGE = {
    lowerBound: `date('now', 'weekday 0')`,
    upperBound: `date('now', 'weekday 0', '+6 days')`,
};
const THIS_MONTH_RANGE = {
    lowerBound: `date('now', 'start of month')`,
    upperBound: `date('now', 'start of month', '1 month', '0 day')`,
};
const LAST_MONTH_RANGE = {
    lowerBound: `date('now', 'start of month', '-1 month')`,
    upperBound: `date('now', 'start of month', '0 day')`,
};
const NEXT_MONTH_RANGE = {
    lowerBound: `date('now', 'start of month', '+1 month')`,
    // -1 day is the day before the end of the month
    // this needs to be 0 day
    upperBound: `date('now', 'start of month', '+2 month', '0 day')`,
};
const THIS_QUARTER_RANGE = () => {
    return computeQuarterDateRange(DateLiteral.THIS_QUARTER);
};
const LAST_QUARTER_RANGE = () => {
    return computeQuarterDateRange(DateLiteral.LAST_QUARTER);
};
const NEXT_QUARTER_RANGE = () => {
    return computeQuarterDateRange(DateLiteral.NEXT_QUARTER);
};
const THIS_YEAR_RANGE = {
    lowerBound: `date('now', 'start of year')`,
    upperBound: `date('now', 'start of month', '1 year', '-1 day')`,
};
const LAST_YEAR_RANGE = {
    lowerBound: `date('now', 'start of year', '-1 year')`,
    upperBound: `date('now', 'start of year', '-1 day')`,
};
const NEXT_YEAR_RANGE = {
    lowerBound: `date('now', 'start of year', '+1 year')`,
    upperBound: `date('now', 'start of year', '+2 years', '-1 day')`,
};
const LAST_90_DAYS_RANGE = {
    lowerBound: `date('now', '-90 days')`,
    upperBound: `date('now')`,
};
const NEXT_90_DAYS_RANGE = {
    lowerBound: `date('now', '+1 day')`,
    upperBound: `date('now', '+90 days')`,
};
// querter start date in UTC
function quarterStart(date) {
    let month = Math.floor(date.getUTCMonth() / 3) * 3 + 1;
    const year = date.getUTCFullYear();
    return month === 10 ? `${year}-${month}-01` : `${year}-0${month}-01`;
}
function computeQuarterDateRange(enumType) {
    const curQuarterStart = quarterStart(new Date());
    switch (enumType) {
        case DateLiteral.LAST_QUARTER:
            return {
                lowerBound: `date('${curQuarterStart}', '-3 months')`,
                // -1 days would set the day before the end of the quarter
                // 0 days sets it to the end of the month
                upperBound: `date('${curQuarterStart}', '0 days')`,
            };
        case DateLiteral.THIS_QUARTER:
            return {
                lowerBound: `date('${curQuarterStart}')`,
                // -1 days would set the day before the end of the quarter
                // 0 days sets it to the end of the month
                upperBound: `date('${curQuarterStart}', '3 months', '0 days')`,
            };
        case DateLiteral.NEXT_QUARTER:
            return {
                lowerBound: `date('${curQuarterStart}', '+3 months')`,
                // -1 days would set the day before the end of the quarter
                // 0 days sets it to the end of the month
                upperBound: `date('${curQuarterStart}', '+6 months', '0 days')`,
            };
    }
}
function dateLiteralRange(literal) {
    switch (literal) {
        case DateLiteral.NEXT_YEAR: {
            return NEXT_YEAR_RANGE;
        }
        case DateLiteral.LAST_WEEK: {
            return LAST_WEEK_RANGE;
        }
        case DateLiteral.THIS_MONTH: {
            return THIS_MONTH_RANGE;
        }
        case DateLiteral.THIS_QUARTER: {
            return THIS_QUARTER_RANGE();
        }
        case DateLiteral.LAST_MONTH: {
            return LAST_MONTH_RANGE;
        }
        case DateLiteral.LAST_90_DAYS: {
            return LAST_90_DAYS_RANGE;
        }
        case DateLiteral.LAST_QUARTER: {
            return LAST_QUARTER_RANGE();
        }
        case DateLiteral.THIS_YEAR: {
            return THIS_YEAR_RANGE;
        }
        case DateLiteral.NEXT_QUARTER: {
            return NEXT_QUARTER_RANGE();
        }
        case DateLiteral.NEXT_MONTH: {
            return NEXT_MONTH_RANGE;
        }
        case DateLiteral.TOMORROW: {
            return TOMORROW_RANGE;
        }
        case DateLiteral.TODAY: {
            return TODAY_RANGE;
        }
        case DateLiteral.LAST_YEAR: {
            return LAST_YEAR_RANGE;
        }
        case DateLiteral.NEXT_90_DAYS: {
            return NEXT_90_DAYS_RANGE;
        }
        case DateLiteral.THIS_WEEK: {
            return THIS_WEEK_RANGE;
        }
        case DateLiteral.YESTERDAY: {
            return YESTERDAY_RANGE;
        }
        case DateLiteral.NEXT_WEEK: {
            return NEXT_WEEK_RANGE;
        }
        default: {
            // eslint-disable-next-line @salesforce/lds/no-error-in-production
            throw new Error(`${literal} is currently not supported in graphql evaluation`);
        }
    }
}
function dateLiteral(literal, op, field, alias) {
    let operator = op;
    let value;
    let bindings = [];
    const { lowerBound, upperBound } = dateLiteralRange(literal);
    switch (operator) {
        case '=': {
            operator = 'BETWEEN';
            value = `${lowerBound} AND ${upperBound}`;
            break;
        }
        case '!=': {
            operator = 'NOT BETWEEN';
            value = `${lowerBound} AND ${upperBound}`;
            break;
        }
        case '<=':
        case '>': {
            value = upperBound;
            break;
        }
        case '<':
        case '>=': {
            value = lowerBound;
            break;
        }
        default: {
            // eslint-disable-next-line @salesforce/lds/no-error-in-production
            throw new Error(`Operator ${operator} is not supported in Date Literals for graphql evaluation`);
        }
    }
    return {
        leftPath: `$.fields.${field.apiName}.value`,
        operator,
        value,
        bindings,
        type: PredicateType.single,
        alias,
        dataType: field.dataType,
        isCaseSensitive: false,
    };
}

var DateRange;
(function (DateRange) {
    DateRange["last_n_days"] = "last_n_days";
    DateRange["next_n_days"] = "next_n_days";
    DateRange["n_days_ago"] = "n_days_ago";
    DateRange["last_n_weeks"] = "last_n_weeks";
    DateRange["next_n_weeks"] = "next_n_weeks";
    DateRange["n_weeks_ago"] = "n_weeks_ago";
    DateRange["last_n_months"] = "last_n_months";
    DateRange["next_n_months"] = "next_n_months";
    DateRange["n_months_ago"] = "n_months_ago";
    DateRange["last_n_quarters"] = "last_n_quarters";
    DateRange["next_n_quarters"] = "next_n_quarters";
    DateRange["n_quarters_ago"] = "n_quarters_ago";
    DateRange["last_n_fiscal_quarters"] = "last_n_fiscal_quarters";
    DateRange["next_n_fiscal_quarters"] = "next_n_fiscal_quarters";
    DateRange["last_n_years"] = "last_n_years";
    DateRange["next_n_years"] = "next_n_years";
    DateRange["n_years_ago"] = "n_years_ago";
    DateRange["last_n_fiscal_years"] = "last_n_fiscal_years";
    DateRange["next_n_fiscal_years"] = "next_n_fiscal_years";
})(DateRange || (DateRange = {}));
function dateTimePredicate(input, operator, field, alias) {
    const { value, literal, range } = input;
    const predicate = {
        leftPath: `$.fields.${field.apiName}.value`,
        operator,
        value,
        alias,
        dataType: field.dataType,
        type: PredicateType.single,
        isCaseSensitive: false,
    };
    if (value !== undefined) {
        if (value === null) {
            switch (predicate.operator) {
                case '=':
                    predicate.operator = 'IS';
                    break;
                case '!=':
                    predicate.operator = 'IS NOT';
                    break;
                default:
                    // eslint-disable-next-line @salesforce/lds/no-error-in-production
                    throw new Error(`Unsupported operator ${predicate.operator} with null value operand`);
            }
            predicate.value = 'NULL';
        }
        else {
            const dateFn = field.dataType === 'Date' ? 'date' : 'datetime';
            predicate.value = `${dateFn}(?)`;
            predicate.bindings = [value];
        }
        return predicate;
    }
    else if (literal !== undefined) {
        const isAvailableLiteral = values$1(DateLiteral).includes(literal);
        // eslint-disable-next-line @salesforce/lds/no-error-in-production
        if (!isAvailableLiteral)
            throw new Error(`${literal} is not a valid DateLiteral`);
        return dateLiteral(literal, operator, field, alias);
    }
    else if (range !== undefined) {
        return dateTimeRange(range, operator, field, alias);
    }
    // eslint-disable-next-line @salesforce/lds/no-error-in-production
    throw new Error(`Where filter ${stringify(input)} is not supported`);
}
function dateTimeRange(input, op, field, alias) {
    const dateFunction = field.dataType === 'DateTime' ? 'datetime' : 'date';
    const key = keys$1(input)[0];
    let operator = op;
    if (operator === '=')
        operator = 'BETWEEN';
    if (operator === '!=')
        operator = 'NOT BETWEEN';
    let value;
    let bindings = [];
    const { range, binding } = dateRangesFrom(key, input[key], dateFunction);
    switch (operator) {
        case '<':
        case '>=':
            value = range.lower;
            bindings = [binding.lower];
            break;
        case '>':
        case '<=':
            value = range.upper;
            bindings = [binding.upper];
            break;
        default:
            value = `${range.lower} AND ${range.upper}`;
            bindings = [binding.lower, binding.upper];
            break;
    }
    return {
        leftPath: `$.fields.${field.apiName}.value`,
        operator,
        value,
        bindings,
        type: PredicateType.single,
        alias,
        dataType: field.dataType,
    };
}
// relative date ranges defined at https://help.salesforce.com/s/articleView?id=sf.filter_dates_relative.htm&type=5
function dateRangesFrom(dateRange, input, dateFunction) {
    // use 'start of day' to ensure the timestamp is 00:00:00 for date time values
    switch (dateRange) {
        // next_n_days range starts at tomorrow through n days
        case DateRange.next_n_days: {
            return {
                binding: {
                    upper: `${input} days`,
                    lower: '1 days',
                },
                range: {
                    upper: `${dateFunction}('now', 'start of day', ?)`,
                    lower: `${dateFunction}('now', 'start of day', ?)`,
                },
            };
        }
        // last_n_days starts -(n) days until today at the current second
        case DateRange.last_n_days: {
            return {
                binding: {
                    upper: `0 days`,
                    lower: `-${input} days`,
                },
                range: {
                    upper: `${dateFunction}('now', ?)`,
                    lower: `${dateFunction}('now', 'start of day', ?)`,
                },
            };
        }
        // n_days_ago
        // starts 12:00:00 am -(n) days from today
        // ends 11:59:99 pm -(n) days from today
        case DateRange.n_days_ago: {
            return {
                binding: {
                    upper: `${-input + 1} days`,
                    lower: `-${input} days`,
                },
                range: {
                    upper: `${dateFunction}('now', 'start of day', ?, '-1 seconds')`,
                    lower: `${dateFunction}('now', 'start of day', ?)`,
                },
            };
        }
        // next_n_weeks
        // starts at 00:00:00 the first day of the week n weeks after the current week
        // ends 23:59:59 of the last day of week before the current
        case DateRange.next_n_weeks: {
            const isStartOfWeek = isTodayStartOfWeek();
            return {
                binding: {
                    // weekday FIRST_DAY_OF_WEEK goes to the next weekday of FIRST_DAY_OF_WEEK if not that current day
                    // if that current day then it will be that day
                    // so for the upper bound if today is the start of the week we need to add 7 more days to the value to get next week as the start
                    upper: `${input * 7 + (isStartOfWeek ? 7 : 0)} days`,
                    //lower bound starts out 00:00:00 of next week
                    lower: `${isStartOfWeek ? 7 : 0} days`,
                },
                range: {
                    upper: `${dateFunction}('now', 'weekday ${FIRST_DAY_OF_WEEK}', 'start of day', ?, '-1 seconds')`,
                    lower: `${dateFunction}('now', 'weekday ${FIRST_DAY_OF_WEEK}', 'start of day', ?)`,
                },
            };
        }
        // last_n_weeks
        // starts at 00:00:00 of day 0 n weeks before the current week
        // ends 23:59:59 at last day of the week before the current week
        case DateRange.last_n_weeks: {
            const isStartOfWeek = isTodayStartOfWeek();
            return {
                binding: {
                    // ending on at 23:59:59 of the week before the current week
                    // -7 more days if today not start of week
                    upper: `${-(isStartOfWeek ? 0 : 7)} days`,
                    // starting on 00:00:00 of n * weeks before current week
                    // -7 more days if today not the start of the week because weekday FIRST_DAY_OF_WEEK puts us next week day 0
                    lower: `${-input * 7 - (isStartOfWeek ? 0 : 7)} days`,
                },
                range: {
                    upper: `${dateFunction}('now', 'weekday ${FIRST_DAY_OF_WEEK}', 'start of day', ?, '-1 seconds')`,
                    lower: `${dateFunction}('now', 'weekday ${FIRST_DAY_OF_WEEK}', 'start of day', ?)`,
                },
            };
        }
        // n_weeks_ago
        // starts 00:00:00 on day 0 of n weeks before the current week
        // ends 7 days after the start
        case DateRange.n_weeks_ago: {
            const isStartOfWeek = isTodayStartOfWeek();
            return {
                binding: {
                    // end of week n weeks before current week
                    // if today is start of the week then we need to -7 from the days given by n
                    // so add 7 to remove a week from -(n)
                    upper: `${-input * 7 + (isStartOfWeek ? 7 : 0)} days`,
                    // start of the week n weeks from the current week
                    // if today is the start of the week then -(n) will do to be the previous week, but
                    // if today is not the start of the week we need to go back 1 more week since weekday FIRST_DAY_OF_WEEK will put us next week
                    lower: `${-input * 7 - (isStartOfWeek ? 0 : 7)} days`,
                },
                range: {
                    upper: `${dateFunction}('now', 'weekday ${FIRST_DAY_OF_WEEK}', 'start of day', ?, '-1 seconds')`,
                    lower: `${dateFunction}('now', 'weekday ${FIRST_DAY_OF_WEEK}', 'start of day', ?)`,
                },
            };
        }
        // next_n_months
        // starts the first day of the next month at 00:00:00 and
        // ends end of the nth month 23:59:59
        case DateRange.next_n_months: {
            return {
                binding: {
                    upper: `${input + 1} months`,
                    lower: `1 months`,
                },
                range: {
                    upper: `${dateFunction}('now', 'start of month', ?, '-1 seconds')`,
                    lower: `${dateFunction}('now', 'start of month', ?)`,
                },
            };
        }
        // last_n_months
        // starts at 00:00:00 first day of n months before the current month and
        // ends at 23:59:59 the last day of the month before the current month
        case DateRange.last_n_months: {
            return {
                binding: {
                    upper: `0 months`,
                    lower: `${-input} months`,
                },
                range: {
                    upper: `${dateFunction}('now', 'start of month', ?, '-1 seconds')`,
                    lower: `${dateFunction}('now', 'start of month', ?)`,
                },
            };
        }
        // n_months_ago
        // starts at the first day of the month 00:00:00 n months ago from now
        // ends at 23:59:59 of n months ago
        case DateRange.n_months_ago: {
            return {
                binding: {
                    // need to go 1 less month back then -1 seconds to get the end of it
                    upper: `${-input + 1} months`,
                    lower: `-${input} months`,
                },
                range: {
                    upper: `${dateFunction}('now', 'start of month', ?, '-1 seconds')`,
                    lower: `${dateFunction}('now', 'start of month', ?)`,
                },
            };
        }
        // next_n_years
        // starts 00:00:00 Jan 1st the year after the current year
        // ends Dec 31 23:59:59 of the nth year
        case DateRange.next_n_years: {
            return {
                binding: {
                    upper: `+${input + 1} years`,
                    lower: `+1 year`,
                },
                range: {
                    upper: `${dateFunction}('now', 'start of year', ?, '-1 seconds')`,
                    lower: `${dateFunction}('now', 'start of year', ?)`,
                },
            };
        }
        // last_n_years starts
        // starts 00:00:00 Jan 1 n+1 year ago and
        // ends dec 31 23:59:59 of the year before the current
        case DateRange.last_n_years: {
            return {
                binding: {
                    upper: `-1 seconds`,
                    lower: `-${input + 1} years`,
                },
                range: {
                    upper: `${dateFunction}('now', 'start of year', ?)`,
                    lower: `${dateFunction}('now', 'start of year', ?)`,
                },
            };
        }
        // n_years_ago
        // starts 00:00:00 Jan 1 of n years before the current year and
        // ends Dec 31 23:59:59 of that year
        case DateRange.n_years_ago: {
            return {
                binding: {
                    upper: `-${input - 1} years`,
                    lower: `-${input} years`,
                },
                range: {
                    upper: `${dateFunction}('now', 'start of year', ?, '-1 seconds')`,
                    lower: `${dateFunction}('now', 'start of year', ?)`,
                },
            };
        }
        // next_n_quarters
        // starts 00:00:00 first day of the quarter after the current quarter
        // ends 23:59:59 of the last day of n quarters in the future
        case DateRange.next_n_quarters: {
            const currentQuarterString = quarterStart(new Date());
            return {
                binding: {
                    // add another quarter input to -1 seconds to get the end of the last day of the quarter
                    upper: `${(input + 1) * 3} months`,
                    lower: `3 months`,
                },
                range: {
                    upper: `${dateFunction}('${currentQuarterString}', 'start of day', ?, '-1 seconds')`,
                    lower: `${dateFunction}('${currentQuarterString}', 'start of day', ?)`,
                },
            };
        }
        // last_n_quarters
        // starts 00:00:00 the first day of n quarters ago
        // end 23:59:59 the last day of the quarter before the current quarter
        case DateRange.last_n_quarters: {
            const currentQuarterString = quarterStart(new Date());
            return {
                binding: {
                    upper: `-1 seconds`,
                    lower: `-${input * 3} months`,
                },
                range: {
                    upper: `${dateFunction}('${currentQuarterString}', 'start of day', ?)`,
                    lower: `${dateFunction}('${currentQuarterString}', 'start of day', ?)`,
                },
            };
        }
        // n_quarters_ago
        // starts 00:00:00 first day of the quarter n quarters before the current
        // ends 23:59:59 last day of the same quarter
        case DateRange.n_quarters_ago: {
            const currentQuarterString = quarterStart(new Date());
            return {
                binding: {
                    // minus 1 quarter to be able to -1 seconds to get the end of the nth quarter
                    upper: `-${(input - 1) * 3} months`,
                    lower: `-${input * 3} months`,
                },
                range: {
                    upper: `${dateFunction}('${currentQuarterString}', 'start of day', ?, '-1 seconds')`,
                    lower: `${dateFunction}('${currentQuarterString}', 'start of day', ?)`,
                },
            };
        }
        default:
            // eslint-disable-next-line @salesforce/lds/no-error-in-production
            throw new Error(`DateRangeInput ${dateRange} is not supported in local evalutation`);
    }
}
function isTodayStartOfWeek() {
    return new Date().getDay() === FIRST_DAY_OF_WEEK;
}

const JSON_EXTRACT_PATH_INGESTION_TIMESTAMP = '$.ingestionTimestamp';
const JSON_EXTRACT_PATH_INGESTION_APINAME = '$.apiName';
const JSON_EXTRACT_PATH_DRAFTS = '$.drafts';

const MultiPickListValueSeparator = ';';
function filterToPredicates(where, recordType, alias, objectInfoMap, joins, draftFunctions) {
    if (!where)
        return [];
    let predicates = [];
    const fields = keys$1(where);
    for (const field of fields) {
        if (field === 'and' || field === 'or') {
            predicates.push(processCompoundPredicate(field, where[field], recordType, alias, objectInfoMap, joins));
        }
        if (field === 'not') {
            // get the predicates that are one level below the not field
            const children = filterToPredicates(where[field], recordType, alias, objectInfoMap, joins);
            // wrap a not predicate around all the child/children predicate
            const notPredicates = children.map((predicate) => {
                return {
                    type: PredicateType.not,
                    operator: 'not',
                    children: [predicate],
                };
            });
            predicates = predicates.concat(notPredicates);
        }
        const fieldInfo = findFieldInfo(objectInfoMap[recordType], field);
        if (!fieldInfo) {
            continue;
        }
        // handles spanning field filter. For example, for 'OwnerId' field within 'Account' objectInfo, both 'OwnerId' and 'Owner' can be
        // treated as reference type, while 'Owner' field spanns since it equals to 'relationshipName'
        if (fieldInfo.dataType === 'Reference' && fieldInfo.relationshipName === field) {
            const path = {
                leftPath: `$.fields.${fieldInfo.apiName}.value`,
                rightPath: `$.id`,
            };
            const childAlias = `${alias}_${field}`;
            // Join can happen on polymorphic field too which has more than 1 `referenceToInfo`. If 'apiName' is undefined, it indicates a Join on polymorphic field.
            const join = {
                type: 'LEFT',
                alias: childAlias,
                to: alias,
                conditions: [path],
                apiName: fieldInfo.referenceToInfos.length === 1
                    ? fieldInfo.referenceToInfos[0].apiName
                    : undefined,
            };
            joins.push(join);
            // 'RelationshipName' can be different from 'apiName'. For example, 'OwnerId' field
            // in 'Account' has the 'Owner' as the 'RelationshipName', the api name in 'referenceToInfos' is 'User'
            if (fieldInfo.referenceToInfos.length === 1) {
                const childRecordType = fieldInfo.referenceToInfos[0].apiName;
                predicates.push(...filterToPredicates(where[field], childRecordType, childAlias, objectInfoMap, joins));
            }
            else {
                // @W-12618378 polymorphic query sometimes does not work as expected on server. The GQL on certain entities could fail.
                const entityNames = keys$1(where[field]);
                const polyPredicatesGroups = entityNames
                    .filter((entityName) => fieldInfo.referenceToInfos.some((referenceInfo) => referenceInfo.apiName === entityName))
                    .map((entityName) => {
                    return [
                        {
                            alias: childAlias,
                            leftPath: JSON_EXTRACT_PATH_INGESTION_APINAME,
                            operator: '=',
                            value: entityName,
                            dataType: 'String',
                            type: PredicateType.single,
                            isCaseSensitive: true,
                        },
                        ...filterToPredicates(where[field][entityName], entityName, childAlias, objectInfoMap, joins),
                    ];
                });
                if (polyPredicatesGroups.length === 1) {
                    predicates.push(...polyPredicatesGroups[0]);
                }
                else {
                    // if polymorphic field has more than 1 entity like `{Owner: { User: {...} Group: {...}}}`, precidates for each entity(`User` and `Group`) should have
                    // `and` operation. while at the base polymorphic field level (`Owner`), the group of compound predicates for each entity should have `or` operation.
                    const polyCompoundPredicates = polyPredicatesGroups.map((polyEntityPredicates) => transformCompoundPredicate('and', polyEntityPredicates));
                    predicates.push(transformCompoundPredicate('or', polyCompoundPredicates));
                }
            }
        }
        else {
            //`field` match the filedInfo's apiName
            for (const [op, value] of entries(where[field])) {
                const operator = operatorToSql(op);
                /**
                Two types ID processing might be needed. Draft ID swapping is optional, which depends on DraftFunctions existence.
                 1. Coercing is needed when query is executed against local db if field is Id field of entity or a reference to an entity. It is OK to use 15 char ID to query agaist server,
                    so there is no need to coerce the ID in injected AST which is used by luvio to do the network request when necessary.
                 2. Draft ID swapping is needed when adapter rebuild is triggered. Draft ID swapping also happens in field injection phase when adapter call is invoked. The injected AST node has
                 the canonical ID and is passed to Luvio
                */
                const idProcessingNeeded = isIDValueField(fieldInfo);
                const makePredicate = fieldInfo.dataType === 'MultiPicklist'
                    ? createMultiPicklistPredicate
                    : createSinglePredicate;
                const predicate = makePredicate(idProcessingNeeded ? sanitizePredicateIDValue(value, draftFunctions) : value, operator, fieldInfo, alias);
                // for the case where we have a not equals predicate that's value is not null
                // we need to compound into an or statement to also inlcude the null values for that field
                if (operator === '!=' && typeof value === 'string') {
                    // create a is null predicate
                    const isNullPredicate = createSinglePredicate(null, '=', fieldInfo, alias);
                    // compound the predicates into an or predicate
                    predicates.push(transformCompoundPredicate('or', [predicate, isNullPredicate]));
                }
                else {
                    predicates.push(predicate);
                }
            }
        }
    }
    return predicates;
}
/**
 *
 * @param fieldInfo
 * @returns true if field is of IDValue type
 */
function isIDValueField(fieldInfo) {
    return fieldInfo.apiName === 'Id' || fieldInfo.referenceToInfos.length > 0;
}
/**
 * Coerces and swaps ID if it is a draft ID.
 * @param value Predicate value which might be a string value like 'John' or a List like ['Acme', 'John']
 * @param draftFunction optional. If undefined, draft ID swapping does not happen
 * @returns sanitized previdicate value
 */
function sanitizePredicateIDValue(value, draftFunction) {
    if (isArray(value)) {
        return value.map((singleValue) => sanitizePredicateIDValue(singleValue, draftFunction));
    }
    else if (typeof value === 'string' && value === '') {
        return value;
    }
    else {
        const coercedId = getRecordId18(value);
        if (coercedId !== undefined) {
            if (draftFunction !== undefined && draftFunction.isDraftId(coercedId)) {
                return draftFunction.getCanonicalId(coercedId);
            }
            return coercedId;
        }
        return value;
    }
}
function createMultiPicklistPredicate(value, operator, fieldInfo, alias) {
    if (!value || (isArray(value) && value.length === 0)) {
        // eslint-disable-next-line
        throw new Error(`No value specified for MultiPickList filter`);
    }
    // generate single prodicate for = and !=
    if (operator === '=' || operator === '!=') {
        // The raw value could be ';;a; b;;', clean it up to 'a;b'
        const welformatedValue = value
            .toString()
            .split(MultiPickListValueSeparator)
            .map((v) => v.trim())
            .filter((v) => v.length > 0)
            .join(MultiPickListValueSeparator);
        return createSinglePredicate(welformatedValue, operator, fieldInfo, alias);
    }
    // if we have more than 1 value in the array from an includes/excludes we must split the values into
    // a compound OR predicate for includes, AND predicate for excludes. each predicate could a composite
    // if the value in the array is ; separated. sample: ['a:b','c']
    const valueArray = isArray(value) ? value : [value];
    return {
        type: PredicateType.compound,
        operator: operator === 'LIKE' ? 'or' : 'and',
        children: valueArray.map((v) => {
            const splittedValue = v
                .split(MultiPickListValueSeparator)
                .map((v) => v.trim())
                .filter((v) => v.length > 0);
            if (splittedValue.length === 1) {
                return createSinglePredicate(v, operator, fieldInfo, alias);
            }
            else {
                return {
                    type: PredicateType.compound,
                    operator: operator === 'LIKE' ? 'and' : 'or',
                    children: splittedValue.map((vInner) => createSinglePredicate(vInner, operator, fieldInfo, alias)),
                };
            }
        }),
    };
}
function createSinglePredicate(val, operator, field, alias) {
    let value = val;
    if (field.dataType === 'MultiPicklist') {
        value = multiPicklistToSql(operator, val);
    }
    else if (typeof value !== 'object' &&
        !isArray(value) &&
        typeof value !== 'number' &&
        (typeof value !== 'string' || typeof value !== 'boolean')) {
        value = value.toString();
    }
    else if (isDateTimeInput(value)) {
        return dateTimePredicate(value, operator, field, alias);
    }
    let leftPath = `$.fields.${field.apiName}.value`;
    if (field.apiName === 'Id') {
        leftPath = '$.id';
    }
    else if (field.apiName === 'weakEtag') {
        leftPath = '$.weakEtag';
    }
    else if (field.apiName === 'RecordTypeId') {
        leftPath = '$.recordTypeId';
    }
    return {
        alias,
        leftPath,
        operator,
        value,
        dataType: field.dataType,
        type: PredicateType.single,
        isCaseSensitive: isIDField(field),
    };
}
function isIDField(field) {
    return field.apiName === 'Id' || field.referenceToInfos.length > 0;
}
function multiPicklistToSql(operator, value) {
    // Individual multipicklist terms that delimited by semicolon are stored server-side
    // as lexically sorted strings and treated like logical ANDs. We can approximate this
    // behavior in SQL with wildcarded `LIKE` SQL operators.  Terms with no delimiter can
    // be treated as string literals.  Multiple terms are logically OR'd together to
    // match the behavior described in SOQL documentation (https://sfdc.co/c9j0r)
    // To make sure the match is safe for includes/excludes. the value is prefix and
    // suffix with ';', like 'abc' to '%;abc;%'.  raw value for eq and ne.
    if (operator === 'LIKE' || operator === 'NOT LIKE') {
        return `%${MultiPickListValueSeparator}${value}${MultiPickListValueSeparator}%`;
    }
    else {
        return value;
    }
}
// handles 'SinglePredicate' | 'CompoundPredicate'|  'NotPredicate' conversion to SQLite
function predicateToSQL(predicate, defaultAlias, isChildNotPredicate = false) {
    if (isSinglePredicate(predicate)) {
        return singlePredicateToSql(predicate, defaultAlias, isChildNotPredicate);
    }
    else if (isNotPredicate(predicate)) {
        return notPredicateToSql(predicate, defaultAlias);
    }
    else {
        return flattenSqlAndBindings(predicatesToSqls(predicate.children, defaultAlias, isChildNotPredicate), predicate.operator);
    }
}
/**
 * Takes an array of predicates turned to sql and flattens them into one statement string
 * with combined bindings
 * @param sqlAndBindings
 * @param operator
 * @returns
 */
function flattenSqlAndBindings(sqlAndBindings, operator) {
    const joiner = operator === undefined ? ' ' : ` ${operator} `.toUpperCase();
    const sql = sqlAndBindings.map((sqlAndBinding) => sqlAndBinding.sql).join(joiner);
    const binding = sqlAndBindings
        .map((sqlAndBinding) => sqlAndBinding.binding)
        .reduce(flatten, []);
    return { sql, binding };
}
function inOrNotInValuesIncludeNull(predicate) {
    const isInOrNin = predicate.operator === 'IN' || predicate.operator === 'NOT IN';
    const nullValueExists = isArray(predicate.value)
        ? predicate.value.some((v) => v === null)
        : predicate.value === null;
    return isInOrNin && nullValueExists;
}
/**
 * Takes a single predicate and appends AND IS NOT NULL to the end of it.
 * So when we wrap a NOT around the predicate it will include non-null values.
 * @param predicate
 * @param defaultAlias
 * @returns
 */
function singleNotPredicateToIncludeNulls(predicate, defaultAlias) {
    // if we are checking a value of null then there is no need to append on and is not null
    // also do not add to in/nin predicates
    if (predicate.value === null || inOrNotInValuesIncludeNull(predicate)) {
        return predicateToSQL(predicate, defaultAlias);
    }
    const { sql, binding } = predicateToSQL(predicate, defaultAlias);
    const notNull = { ...predicate, operator: 'IS NOT', value: null };
    const notNullSql = predicateToSQL(notNull, defaultAlias);
    return {
        sql: `${sql} AND ${notNullSql.sql}`,
        binding: binding.concat(notNullSql.binding),
    };
}
function notPredicateToSql(predicate, defaultAlias) {
    const children = predicate.children.map((child) => {
        if (isNotPredicate(child)) {
            // eslint-disable-next-line @salesforce/lds/no-error-in-production
            throw new Error('A not predicate cannot be a child of a not query');
        }
        if (isSinglePredicate(child)) {
            const { sql, binding } = singleNotPredicateToIncludeNulls(child, defaultAlias);
            return { sql: `NOT ( ${sql} )`, binding };
        }
        else {
            const compound = child.children.map((c) => {
                if (isSinglePredicate(c)) {
                    const { sql, binding } = singleNotPredicateToIncludeNulls(c, defaultAlias);
                    return { sql: `( ${sql} )`, binding };
                }
                else {
                    return predicateToSQL(c, defaultAlias, true);
                }
            });
            const { sql, binding } = flattenSqlAndBindings(compound, child.operator);
            return { sql: `( NOT ( ${sql} ) )`, binding };
        }
    });
    return flattenSqlAndBindings(children);
}
/**
 * Used to build a SinglePredicate into a sql statement with bindings
 * @param predicate
 * @param defaultAlias
 * @returns
 */
function singlePredicateToSql(predicate, defaultAlias, isChildNotPredicate = false) {
    //if a child of a not predicate than we need to include nulls
    if (isChildNotPredicate) {
        return singleNotPredicateToIncludeNulls(predicate, defaultAlias);
    }
    // handles different type of filter predicate like 'date', 'picklist', 'time', 'percent'
    let { alias, leftPath, operator, value, dataType, isCaseSensitive } = predicate;
    let binding = [];
    let boundValue = extractPredicateValue(value, dataType);
    // Handles boolean type field
    if (dataType === 'Boolean') {
        if (boundValue === null) {
            if (operator === '=') {
                operator = 'IS';
            }
            boundValue = 'NULL';
        }
        else if (boundValue === true) {
            boundValue = 1;
        }
        else {
            boundValue = 0;
        }
    }
    if (boundValue === null) {
        if (operator === '=') {
            operator = 'IS';
        }
        else if (operator === '!=') {
            operator = 'IS NOT';
        }
    }
    alias = !alias ? defaultAlias : alias;
    let sql;
    if (predicate.bindings !== undefined) {
        if (predicate.dataType !== undefined &&
            (predicate.dataType === 'DateTime' || predicate.dataType === 'Date')) {
            const dateFunction = predicate.dataType === 'DateTime' ? 'datetime' : 'date';
            sql = `${dateFunction}(json_extract("${alias}".data, '${leftPath}')) ${operator} ${predicate.value}`;
        }
        else {
            sql = `json_extract("${alias}".data, '${leftPath}') ${operator} ${predicate.value}`;
        }
        binding = binding.concat(predicate.bindings);
    }
    else {
        if (operator === 'IN' || operator === 'NOT IN') {
            const { sql: questionSql, binding: valueBinding, includesNull, } = handleExtractedPredicateValue(boundValue, true);
            // If an explicit collating sequence is required on an IN operator it should be applied to the left operand,
            // like this: "x COLLATE NOCASE IN (y,z, ...)".
            const nullCheck = `json_extract("${alias}".data, '${leftPath}') ${operator === 'IN' ? 'IS' : 'IS NOT'} ?`;
            sql = `json_extract("${alias}".data, '${leftPath}')${isCaseSensitive === true ? '' : ` COLLATE NOCASE`} ${operator} ${questionSql}`;
            binding.push(...valueBinding);
            if (includesNull) {
                sql = `(${sql} OR ${nullCheck})`;
                binding.push(null);
            }
        }
        else {
            const { sql: questionSql, binding: valueBinding } = handleExtractedPredicateValue(boundValue, false);
            // SQLite is case sensitive by default, SOQL is case in-sensitive by default
            // For pick list includes or excludeds, prefix and suffix the field value with ';' to guarantee the query accuracy.
            if (dataType === 'MultiPicklist' && (operator === 'LIKE' || operator === 'NOT LIKE')) {
                sql = `'${MultiPickListValueSeparator}' || json_extract("${alias}".data, '${leftPath}') || '${MultiPickListValueSeparator}' ${operator} ${questionSql} COLLATE NOCASE`;
            }
            else {
                sql = `json_extract("${alias}".data, '${leftPath}') ${operator} ${questionSql}${isCaseSensitive === true ? '' : ` COLLATE NOCASE`}`;
            }
            binding.push(...valueBinding);
        }
    }
    return { sql, binding };
}
// for one value, return { sql: '?', bindings:'xxx'}. for multiple value, return { sql: '(?, ?, ?)', binding: ['xxx','yyy','zzz'] }
function handleExtractedPredicateValue(boundValue, checkForNull) {
    let questionSql = '?';
    let binding = [];
    let includesNull = false;
    if (isArray(boundValue)) {
        const removeNull = boundValue.filter((v) => v !== null);
        includesNull = removeNull.length !== boundValue.length;
        // construct (?, ?)
        questionSql = `(${new Array(checkForNull ? removeNull.length : boundValue.length)
            .fill('?')
            .join(', ')})`;
        binding = binding.concat(...(checkForNull ? removeNull : boundValue));
    }
    else {
        binding = binding.concat(boundValue);
    }
    return { sql: questionSql, binding, includesNull };
}
function processCompoundPredicate(operator, filters, recordType, alias, objectInfoMap, joins) {
    const predicates = filters
        .map((filter) => filterToPredicates(filter, recordType, alias, objectInfoMap, joins))
        .reduce(flatten);
    return transformCompoundPredicate(operator, predicates);
}
function transformCompoundPredicate(operator, childPredicates) {
    const diffCompoundPredicates = childPredicates
        .filter(isCompoundPredicate)
        .filter((compoundPredicate) => compoundPredicate.operator !== operator);
    // flattens the child compound predicate if its operator is the same of its parent to improve the SQL efficiency.
    const sameCompoundPredicates = childPredicates
        .filter(isCompoundPredicate)
        .filter((compoundPredicate) => compoundPredicate.operator === operator)
        .map((pred) => pred.children)
        .reduce(flatten, []);
    const singlePredicates = childPredicates.filter((p) => isSinglePredicate(p) || isNotPredicate(p));
    const children = [...diffCompoundPredicates, ...sameCompoundPredicates, ...singlePredicates];
    if (children.length === 1) {
        return children[0];
    }
    return {
        type: PredicateType.compound,
        operator,
        children,
    };
}
function predicatesToSqls(predicates, defaultAlias, isChildNotPredicate = false) {
    // (a and (b or c)) is differnt from (a and b or c)
    const wrapit = predicates.length >= 1 && predicates.some(isCompoundPredicate);
    const sqlAndBindingsResult = [];
    if (wrapit) {
        const unwrappedSqlAndBindings = predicates
            .filter(isSinglePredicate)
            .map((predicate) => predicateToSQL(predicate, defaultAlias, isChildNotPredicate));
        if (unwrappedSqlAndBindings && unwrappedSqlAndBindings.length) {
            sqlAndBindingsResult.push(...unwrappedSqlAndBindings);
        }
        const wrappedSqlAndBindings = predicates
            .filter(isCompoundPredicate)
            .map((predicate) => predicateToSQL(predicate, defaultAlias, isChildNotPredicate))
            .map(wrapSqlAndBindings);
        sqlAndBindingsResult.push(...wrappedSqlAndBindings);
    }
    else {
        const sqlAndBindings = predicates.map((predicate) => predicateToSQL(predicate, defaultAlias, isChildNotPredicate));
        sqlAndBindingsResult.push(...sqlAndBindings);
    }
    return sqlAndBindingsResult;
}
function wrapSqlAndBindings(sqlAndBindings) {
    return { sql: `(` + sqlAndBindings.sql + `)`, binding: sqlAndBindings.binding };
}
function operatorToSql(graphqlOperator) {
    switch (graphqlOperator) {
        case 'eq':
            return '=';
        case 'ne':
            return '!=';
        case 'gt':
            return '>';
        case 'gte':
            return '>=';
        case 'lt':
            return '<';
        case 'lte':
            return '<=';
        case 'in':
            return 'IN';
        case 'nin':
            return 'NOT IN';
        case 'like':
        case 'includes':
            return 'LIKE';
        case 'excludes':
            return 'NOT LIKE';
        default:
            // eslint-disable-next-line
            throw new Error(`Unknown GraphQL operator: ${graphqlOperator}`);
    }
}
function extractPredicateValue(value, dataType) {
    if (dataType !== undefined && dataType === 'Boolean') {
        if (typeof value === 'string') {
            return value === 'true';
        }
        if (typeof value === 'boolean')
            return value;
    }
    if (typeof value !== 'object') {
        if (typeof value === 'string' || typeof value === 'number' || typeof value === 'boolean') {
            return value;
        }
        else {
            return value.toString();
        }
    }
    return value;
}

var PredicateType;
(function (PredicateType) {
    PredicateType["compound"] = "compound";
    PredicateType["single"] = "single";
    PredicateType["not"] = "not";
})(PredicateType || (PredicateType = {}));
function buildQuery(config) {
    const joins = buildJoins(config);
    const predicates = buildPredicates(config);
    const orderBy = buildOrderBy(config);
    const staleRecordsSql = excludeStaleRecordsGate.isOpen({ fallback: false })
        ? `AND (
            json_extract("${config.alias}".metadata, '${JSON_EXTRACT_PATH_INGESTION_TIMESTAMP}') >= ?
            OR json_extract("${config.alias}".data, '${JSON_EXTRACT_PATH_DRAFTS}') IS NOT NULL
          )`
        : '';
    const sql = `
        SELECT "${config.alias}".data
        FROM lds_data "${config.alias}" ${joins.sql}
        WHERE "${config.alias}".key like 'UiApi::RecordRepresentation:%'
          AND json_extract("${config.alias}".data, '${JSON_EXTRACT_PATH_INGESTION_APINAME}') = '${config.alias}'
          ${staleRecordsSql}
          ${predicates.sql}
          ${orderBy.sql}
        LIMIT ?
        `
        .split('\n')
        .map((line) => line.trim())
        .join(' ');
    const bindings = [
        // bindings from predicates on joins
        ...joins.bindings,
        ...(excludeStaleRecordsGate.isOpen({ fallback: false }) ? [config.ingestionTimestamp] : []),
        // where clause and parent scope bindings
        ...predicates.bindings,
        // limit binding
        config.limit || 10,
    ];
    return { sql: sql.trim(), bindings };
}
function dedupeJoins(joins) {
    const deduped = {};
    for (const join of joins) {
        deduped[join.alias + join.to] = join;
    }
    return values$1(deduped);
}
function buildJoins(config) {
    let sql = '';
    const bindings = [];
    const orderByJoins = config.orderBy !== undefined ? removeUndefined(config.orderBy.map((x) => x.joins)) : [];
    const allJoins = dedupeJoins([
        ...(config.joins || []),
        ...(orderByJoins.reduce(flatten, []) || []),
    ]);
    if (allJoins.length === 0)
        return { sql, bindings };
    sql = allJoins.reduce((joinAccumulator, join) => {
        let timestampAdded = false;
        const joinConditions = join.conditions.reduce((conditionAccumulator, condition) => {
            let joined_sql;
            const joinMetadataTimestamp = excludeStaleRecordsGate.isOpen({ fallback: false })
                ? ` AND (json_extract("${join.alias}".metadata, '${JSON_EXTRACT_PATH_INGESTION_TIMESTAMP}') >= ? OR json_extract("${join.alias}".data, '${JSON_EXTRACT_PATH_DRAFTS}') IS NOT NULL)`
                : '';
            // predicate on a value, use the newly joined table
            if ('type' in condition) {
                const { sql, binding } = predicateToSQL(condition, join.alias);
                joined_sql = ` AND ${sql}${timestampAdded ? '' : joinMetadataTimestamp}`;
                bindings.push(...binding);
                if (excludeStaleRecordsGate.isOpen({ fallback: false }) &&
                    timestampAdded === false) {
                    bindings.push(config.ingestionTimestamp);
                    timestampAdded = true;
                }
            }
            else {
                // predicate on a path
                const left = ` AND json_extract("${join.to}".data, '${condition.leftPath}')`;
                const right = `json_extract("${join.alias}".data, '${condition.rightPath}')`;
                joined_sql = `${left} = ${right}${timestampAdded ? '' : joinMetadataTimestamp}`;
                if (excludeStaleRecordsGate.isOpen({ fallback: false }) &&
                    timestampAdded === false) {
                    bindings.push(config.ingestionTimestamp);
                    timestampAdded = true;
                }
            }
            conditionAccumulator += joined_sql;
            return conditionAccumulator;
        }, '');
        // 'joinConditions' has the prefix 'AND_' which needs to be removed.
        const formattedJoinCondition = joinConditions.slice(4);
        joinAccumulator += `
            ${join.type} JOIN lds_data "${join.alias}" 
            ON "${join.alias}".key like 'UiApi::RecordRepresentation:%'
                AND${formattedJoinCondition}
        `;
        return joinAccumulator;
    }, '');
    return { sql, bindings };
}
function buildPredicates(config) {
    if (!config.predicates)
        return { sql: '', bindings: [] };
    const sqlAndBindingsResult = predicatesToSqls(config.predicates, config.alias);
    let sql = sqlAndBindingsResult.map((sqlAndBinding) => sqlAndBinding.sql).join(` AND `);
    if (sql && sql.length > 0) {
        sql = `AND ` + sql;
    }
    const bindings = sqlAndBindingsResult
        .map((sqlAndBinding) => sqlAndBinding.binding)
        .reduce(flatten, []);
    return { sql, bindings };
}
function buildOrderBy(config) {
    let sql = '';
    if (!config.orderBy)
        return { sql };
    const orderByStatements = config.orderBy.reduce((accumulator, orderBy) => {
        const sortOrder = orderBy.order !== undefined ? `${orderBy.order}` : ``;
        const sortNulls = orderBy.nulls !== undefined ? ` NULLS ${orderBy.nulls}` : ``;
        const statement = `json_extract("${orderBy.alias}".data, '${orderBy.path}') ${sortOrder} ${sortNulls}`;
        accumulator.push(statement);
        return accumulator;
    }, []);
    sql = `
        ${orderByStatements.length > 0 ? 'ORDER BY' : ''}
        ${orderByStatements.join(', ')}
    `;
    return { sql };
}
function isCompoundPredicate(predicate) {
    return predicate.type === PredicateType.compound;
}
function isSinglePredicate(predicate) {
    return predicate.type === PredicateType.single;
}
function isNotPredicate(predicate) {
    return predicate.type === PredicateType.not;
}
function isDateTimeInput(object) {
    return (object !== undefined &&
        object !== null &&
        typeof object === 'object' &&
        (object.value !== undefined || object.literal !== undefined || object.range !== undefined));
}
const PARENT_RELATIONSHIP = 'parentRelationship';
function depth(json, currentLevel = 0) {
    if (typeof json !== 'object') {
        return currentLevel;
    }
    const keys$1$1 = keys$1(json);
    if (keys$1$1.length === 0)
        return 0;
    const depths = keys$1$1.map((key) => {
        const value = json[key];
        return depth(value, currentLevel + 1);
    });
    return Math.max(...depths);
}
function flatten(previous, current) {
    return previous.concat(current);
}
function findFieldInfo(objectInfo, fieldName) {
    return values$1(objectInfo.fields).find((field) => field.apiName === fieldName ||
        (field.dataType === 'Reference' && field.relationshipName === fieldName));
}
async function readIngestionTimestampForKey(key, query) {
    let ingestionTimestamp = 0;
    const sql = `SELECT json_extract(metadata, '${JSON_EXTRACT_PATH_INGESTION_TIMESTAMP}') FROM lds_data WHERE key IS ?`;
    const results = await query(sql, [key]);
    const [timestamp] = results.rows.map((row) => row[0]);
    if (timestamp !== null) {
        const numericalTimestamp = Number(timestamp);
        if (isNaN(numericalTimestamp)) {
            return ingestionTimestamp;
        }
        ingestionTimestamp = numericalTimestamp;
    }
    return ingestionTimestamp;
}

function findSpanningField(name) {
    return (field) => {
        return (field.apiName === name ||
            (field.dataType === 'Reference' && field.relationshipName === name));
    };
}
function orderByToPredicate(orderBy, recordType, alias, objectInfoMap, joins) {
    let predicates = [];
    if (!orderBy)
        return predicates;
    const isSpanning = depth(orderBy) > 2;
    if (isSpanning) {
        const keys$1$1 = keys$1(orderBy);
        for (let i = 0, len = keys$1$1.length; i < len; i++) {
            const key = keys$1$1[i];
            const parentFields = objectInfoMap[recordType].fields;
            const fieldInfo = values$1(parentFields).find(findSpanningField(key));
            if (fieldInfo && fieldInfo.referenceToInfos.length > 0) {
                const { apiName } = fieldInfo.referenceToInfos[0];
                const parentFieldInfo = values$1(objectInfoMap[recordType].fields).find(findSpanningField(fieldInfo.apiName));
                if (parentFieldInfo !== undefined) {
                    const path = {
                        leftPath: `$.fields.${parentFieldInfo.apiName}.value`,
                        rightPath: '$.id',
                    };
                    const childAlias = `${alias}_${fieldInfo.relationshipName}`;
                    const join = [
                        {
                            type: 'LEFT',
                            alias: childAlias,
                            to: alias,
                            conditions: [path],
                            apiName: apiName,
                        },
                    ];
                    let spanningPredicates = orderByToPredicate(orderBy[key], apiName, childAlias, objectInfoMap, joins ? joins.concat(join) : join);
                    predicates = predicates.concat(spanningPredicates);
                }
            }
        }
    }
    else {
        const keys$1$1 = keys$1(orderBy);
        for (let i = 0, len = keys$1$1.length; i < len; i++) {
            const key = keys$1$1[i];
            if (!objectInfoMap[recordType])
                continue;
            const fieldInfo = objectInfoMap[recordType].fields[key];
            if (fieldInfo === undefined)
                continue;
            const { order, nulls } = orderBy[key];
            const predicate = {
                path: pathForKey(key),
                alias,
                order,
                nulls,
                joins,
            };
            predicates.push(predicate);
        }
    }
    return predicates;
}
function pathForKey(key) {
    switch (key) {
        case 'ApiName':
            return JSON_EXTRACT_PATH_INGESTION_APINAME;
        case 'WeakEtag':
            return '$.weakEtag';
        case 'Id':
            return '$.id';
        default:
            return `$.fields.${key}.value`;
    }
}

function scopeToJoins(scope = '', settings) {
    if (scope !== 'ASSIGNEDTOME')
        return [];
    return [
        {
            alias: 'ServiceAppointment_AssignedResource',
            type: 'INNER',
            to: 'ServiceAppointment',
            conditions: [
                {
                    type: PredicateType.single,
                    alias: 'ServiceAppointment_AssignedResource',
                    leftPath: JSON_EXTRACT_PATH_INGESTION_APINAME,
                    operator: '=',
                    value: 'AssignedResource',
                    dataType: 'String',
                    isCaseSensitive: true,
                },
                {
                    leftPath: '$.id',
                    rightPath: '$.fields.ServiceAppointmentId.value',
                },
            ],
            apiName: 'AssignedResource',
        },
        {
            alias: 'ServiceAppointment_AssignedResource_ServiceResource',
            type: 'INNER',
            to: 'ServiceAppointment_AssignedResource',
            conditions: [
                {
                    type: PredicateType.single,
                    alias: 'ServiceAppointment_AssignedResource_ServiceResource',
                    leftPath: JSON_EXTRACT_PATH_INGESTION_APINAME,
                    operator: '=',
                    value: 'ServiceResource',
                    dataType: 'String',
                    isCaseSensitive: true,
                },
                {
                    leftPath: '$.fields.ServiceResourceId.value',
                    rightPath: '$.id',
                },
                {
                    type: PredicateType.single,
                    alias: 'ServiceAppointment_AssignedResource_ServiceResource',
                    leftPath: '$.fields.RelatedRecordId.value',
                    operator: '=',
                    value: settings.userId,
                    dataType: 'String',
                    isCaseSensitive: true,
                },
            ],
            apiName: 'ServiceResource',
        },
    ];
}
function scopeToPredicates(scope = '', settings) {
    if (scope !== 'MINE')
        return [];
    return [
        {
            type: PredicateType.single,
            leftPath: '$.fields.OwnerId.value',
            operator: '=',
            value: settings.userId,
        },
    ];
}

function noOp() {
    return {};
}
/**
 *
 * @param schema GraphQL Schema generated from ObjectInfos
 * @param polyFields The record types which support polymorphic fields. For example, 'Owner' has the 'Record' interface which is a polymorphic fields. The 'User' and 'Group' are the concrete types which support 'Owner' interface.
 * @returns Updated schema
 */
function addResolversToSchema(schema, polyFields) {
    // Polymorphic field type is 'recordInterface'
    let recordInterface = undefined;
    let baseRecord = undefined;
    // Concrete types for Polymorphic field
    const polyTypes = [];
    for (const type of values$1(schema.getTypeMap())) {
        if (type.name === 'Record') {
            recordInterface = type;
        }
        else if (type.name === 'RecordRepresentation') {
            baseRecord = type;
        }
        if (!isObjectType(type))
            continue;
        // Finds entity record type which supports polymorphism.
        if (polyFields.find((fieldTypeName) => fieldTypeName === type.name) !== undefined) {
            polyTypes.push(type);
        }
        const fields = values$1(type.getFields());
        // initialize the fields of current type with default behavior
        for (const field of fields) {
            field.resolve = defaultFieldResolver;
        }
        if (type.name === 'Query') {
            for (const field of fields) {
                field.resolve = noOp;
            }
        }
        if (type.name === 'UIAPI') {
            for (const field of fields) {
                if (field.name === 'query') {
                    field.resolve = noOp;
                }
                else {
                    const fieldName = field.name;
                    field.resolve = (_obj, _args, { snapshot }) => {
                        if (!snapshot.data)
                            return null;
                        if (!snapshot.data.data)
                            return null;
                        return snapshot.data.data.uiapi[fieldName];
                    };
                }
            }
        }
        if (type.name === 'RecordQuery') {
            // {
            //     uiapi {
            //         query { <-----------|
            //             Account         |
            //             User            |
            //             Opportunity     |
            //         } <-----------------|
            //     }
            // }
            // Fields of the `RecordQuery` type are the record queries for the entity types
            // supported for the org
            for (const recordQuery of fields) {
                recordQuery.resolve = async function recordConnectionResolver(record, args, { query }, info) {
                    const { name: currentFieldName } = recordQuery;
                    let ingestionTimestamp = 0;
                    if (excludeStaleRecordsGate.isOpen({ fallback: false })) {
                        // at our record query we fetch each ingestion time stamp and pass it down to each lower resolver to query against
                        ingestionTimestamp = await fetchIngestionTimeStampFromDatabase(currentFieldName, info, args, query);
                    }
                    // In the SF schema, the relevant arguments are passed into RecordQuery fields, but actually used
                    // down in the edge resolvers.  For this resolver, we can just return what was passed in
                    // to make it available to the next execution step
                    return {
                        parentArgs: args,
                        parentRecord: record,
                        currentFieldName,
                        ingestionTimestamp,
                    };
                };
            }
        }
        if (type.name.endsWith('Connection')) {
            // {
            //     uiapi {
            //         query {
            //             ServiceAppointment @category(name: "recordQuery") {  <---|
            //                 totalCount                                           |
            //                 edges {                                              |
            //                     node {                                           |
            //                         Id                                           |
            //                     }                                                |
            //                 }                                                    |
            //             }   <----------------------------------------------------|
            //         }
            //     }
            // }
            for (const field of fields) {
                switch (field.name) {
                    case 'edges':
                        field.resolve = connectionEdgeResolver;
                        break;
                    default:
                        field.resolve = defaultFieldResolver;
                }
            }
        }
        if (type.name.endsWith('Edge')) {
            // {
            //     uiapi {
            //         query {
            //             ServiceAppointment @category(name: "recordQuery") {
            //                 edges {     <--|
            //                     node {     |
            //                         Id     |
            //                     }          |
            //                 }           <--|
            //             }
            //         }
            //     }
            // }
            for (const field of fields) {
                if (field.name === 'node') {
                    field.resolve = function nodeResolver(obj, _args, { seenRecordIds }) {
                        const { record, ingestionTimestamp } = obj;
                        const recordRepresentation = parse(record);
                        seenRecordIds.add(recordRepresentation.id);
                        return { recordRepresentation, ingestionTimestamp };
                    };
                }
            }
        }
        if (isRecordType(type)) {
            // All the generic named GraphQL Objects that map to Record Queries ("Account", "ServiceAppointment")
            // {
            //     uiapi {
            //         query {
            //             ServiceAppointment @category(name: "recordQuery") {
            //                 edges {
            //                     node {  <--|
            //                         Id     |
            //                     }       <--|
            //                 }
            //             }
            //         }
            //     }
            // }
            for (const field of fields) {
                switch (field.name) {
                    case 'Id':
                        field.resolve = ({ recordRepresentation: record }) => record.id;
                        break;
                    case 'ApiName':
                        field.resolve = ({ recordRepresentation: record }) => record.apiName;
                        break;
                    case 'WeakEtag':
                        field.resolve = ({ recordRepresentation: record }) => record.weakEtag;
                        break;
                    case '_drafts':
                        field.resolve = ({ recordRepresentation: record, }) => {
                            return record.drafts ? record.drafts : null;
                        };
                        break;
                    case 'LastModifiedById':
                        field.resolve = ({ recordRepresentation: record }) => {
                            return record.lastModifiedById
                                ? { value: record.lastModifiedById }
                                : null;
                        };
                        break;
                    case 'LastModifiedDate':
                        field.resolve = ({ recordRepresentation: record }) => {
                            // In UIAPI record reps, LastModifiedDate might be present as a field,
                            // which will include both the value and displayValue
                            if (record.fields['LastModifiedDate']) {
                                return record.fields['LastModifiedDate'];
                            }
                            // If the field is not present, just return the value of the root property
                            return record.lastModifiedDate
                                ? { value: record.lastModifiedDate }
                                : null;
                        };
                        break;
                    case 'SystemModstamp':
                        field.resolve = ({ recordRepresentation: record }) => {
                            return record.systemModstamp ? { value: record.systemModstamp } : null;
                        };
                        break;
                    case 'RecordTypeId':
                        field.resolve = ({ recordRepresentation: record }) => {
                            return record.recordTypeId ? { value: record.recordTypeId } : null;
                        };
                        break;
                    case 'DisplayValue':
                        field.resolve = recordDisplayValueResolver;
                        break;
                    default: {
                        const recordFieldType = schema.getType(field.type.name);
                        // Both concrete record type and 'record' interface type(polymorphic field) uses 'RecordLoader' to resolve.
                        if ((isObjectType(recordFieldType) &&
                            recordFieldType
                                .getInterfaces()
                                .find((iface) => iface.name === 'Record')) ||
                            (recordFieldType && recordFieldType.name === 'Record')) {
                            field.resolve = async function relationResolver(obj, _args, { Record, seenRecordIds }) {
                                const fetchRecordOrNull = async (key) => {
                                    const recordRepresentation = await Record.load(key);
                                    return recordRepresentation !== null
                                        ? {
                                            recordRepresentation,
                                            ingestionTimestamp,
                                        }
                                        : null;
                                };
                                const { recordRepresentation: record, ingestionTimestamp } = obj;
                                let id = undefined;
                                if (field.name === 'RecordType') {
                                    // RecordTypeId has special handling during ingest and is
                                    // not in record.fields, so check for it at the UIAPI root property location
                                    id = record.recordTypeId;
                                }
                                else if (field.name.endsWith('__r')) {
                                    // Custom relationships end in `__r` and the corresponding ID field should be `__c`
                                    let fieldName = field.name.replace('__r', '__c');
                                    id = record.fields[fieldName] && record.fields[fieldName].value;
                                }
                                else {
                                    // Standard relationships are just FieldNameId
                                    let fieldName = field.name + 'Id';
                                    id = record.fields[fieldName] && record.fields[fieldName].value;
                                }
                                if (!id || typeof id !== 'string') {
                                    // possibly field injection did not inject the necessary Id field
                                    // for the relationship, or we found a non-string value.
                                    return null;
                                }
                                seenRecordIds.add(id);
                                return fetchRecordOrNull(id);
                            };
                        }
                        else if (isObjectType(recordFieldType) &&
                            field.type.name.endsWith('Connection')) {
                            // spanning field to a connection
                            field.resolve = async ({ recordRepresentation, ingestionTimestamp }, args, { seenRecordIds }) => {
                                seenRecordIds.add(recordRepresentation.id);
                                const { name: currentFieldName } = field;
                                return {
                                    parentArgs: args,
                                    parentRecord: recordRepresentation,
                                    currentFieldName,
                                    ingestionTimestamp,
                                };
                            };
                        }
                        else {
                            field.resolve = function recordFieldResolver({ recordRepresentation: record, }) {
                                return record.fields[field.name] || null;
                            };
                        }
                    }
                }
            }
        }
    }
    if (recordInterface !== undefined && baseRecord !== undefined) {
        // Applys 'resolveType' of GraphQLInterfaceType to 'Record' interface. Since all the heterogenous types are named as 'apiName', the type with same name as the loaded record 'apiName' property is the type wanted.
        // GraphQL executor would  match InLineFragment' condition with type and keeps the deeper level field resolving going.
        recordInterface.resolveType = function ({ recordRepresentation: value, }) {
            const targetType = polyTypes.find((type) => type.name === value.apiName);
            return targetType === undefined ? baseRecord : targetType;
        };
    }
    return schema;
}
/**
 * Resolver to fetch the DisplayValue of a graphql record.
 * This is taking the main Name field for the record
 * type and using it for the DisplayValue.
 *
 * This allows us to be more performant without needing to call
 * another SQL lookup for the cached graphql object of the RAML record.
 */
function recordDisplayValueResolver({ recordRepresentation }, 
// not used, but gets sent in the resolver
_, { objectInfos }) {
    const { apiName } = recordRepresentation;
    const objectInfo = objectInfos[apiName];
    if (objectInfo !== undefined) {
        const { nameFields, fields } = objectInfo;
        let nameField;
        // use the field Name if it exists or the first field name
        if (nameFields.length !== 0) {
            nameField = nameFields.find((x) => x === 'Name');
            if (nameField === undefined) {
                nameField = nameFields[0];
            }
        }
        if (nameField !== undefined) {
            nameField = fields[nameField].apiName;
            const nameFieldRep = recordRepresentation.fields[nameField];
            if (nameFieldRep !== undefined) {
                return nameFieldRep.value;
            }
        }
    }
    return null;
}
async function connectionEdgeResolver(obj, _args, context) {
    const { parentArgs = {}, parentRecord, currentFieldName, ingestionTimestamp } = obj;
    const { query, objectInfos, draftFunctions } = context;
    let joins = [];
    let alias = currentFieldName;
    let childRelationshipFieldName = undefined;
    if (parentRecord.id) {
        const parentApiName = parentRecord.apiName;
        const parentObjectInfo = objectInfos[parentApiName];
        const childRelationship = parentObjectInfo &&
            parentObjectInfo.childRelationships.find((rel) => rel.relationshipName === currentFieldName);
        // or emit/throw if we want to report it
        if (!childRelationship)
            return [];
        alias = childRelationship.childObjectApiName;
        childRelationshipFieldName = childRelationship.fieldName;
    }
    // Alias starts as entity's ApiName
    const predicates = [
        ...filterToPredicates(parentArgs.where, alias, alias, context.objectInfos, joins, draftFunctions),
        ...scopeToPredicates(parentArgs.scope, context.settings),
        ...childRelationshipToPredicates(childRelationshipFieldName, parentRecord.id),
    ];
    const scopeJoins = scopeToJoins(parentArgs.scope, context.settings);
    joins.push(...scopeJoins);
    // Alias starts as entity's ApiName
    const queryConfig = {
        alias,
        joins,
        predicates,
        orderBy: orderByToPredicate(parentArgs.orderBy, alias, alias, context.objectInfos),
        limit: parentArgs.first,
        ingestionTimestamp,
    };
    const { sql, bindings } = buildQuery(queryConfig);
    const results = await query(sql, bindings);
    //map each sql result with the ingestion timestamp to pass it down a level
    return results.rows
        .map((row) => row[0])
        .map((record) => {
        return {
            record,
            ingestionTimestamp,
        };
    });
}
/**
 * Converts a childRelationship into a predicate
 * @param childRelationshipFieldName Reference ID field name to its parent record. A defined `childRelationshipFieldName` string indicates that a child relationship exists
 *        and a relationship predicate needs to be put into place. For example, `ServiceAppointment` has a child relationship `ServiceResources`, whose entity name is `AssignedResource`.
 *        Once the parent `ServiceAppointment` record comes back, its child connection starts to resolve. Child `AssignedResource` record needs to reference to parent Id using field `ServiceAppointmentId`.
 * @param parentId prarent record Id
 * @returns predicate array consists at most 1 predicate
 */
function childRelationshipToPredicates(childRelationshipFieldName, parentId) {
    const predicates = [];
    if (childRelationshipFieldName !== undefined && parentId !== undefined) {
        predicates.push({
            type: PredicateType.single,
            leftPath: `$.fields.${childRelationshipFieldName}.value`,
            operator: '=',
            value: parentId,
        });
    }
    return predicates;
}
function isRecordType(type) {
    const interfaces = type.getInterfaces();
    return Boolean(interfaces.find((iface) => iface.name === 'Record'));
}
/**
 * Builds the top level record query key based on AST data
 * @param operation
 * @param variables
 * @param argumentNodes
 * @param currentFieldName
 * @returns
 */
function buildKeyStringForRecordQuery(operation, variables, argumentNodes, currentFieldName) {
    const queryKey = buildQueryTypeStringKey({
        luvio: {},
        keyPrefix: 'UiApi',
        schemaName: 'uiapi',
        queryTypeName: 'Query',
        operationNode: operation,
        variables,
        fragmentMap: {},
    });
    const filteredArgumentNodes = assign$1([], argumentNodes).filter((node) => node.name.value !== 'first' && node.name.value !== 'after');
    const argumentString = filteredArgumentNodes.length > 0
        ? '__' + serializeFieldArguments(filteredArgumentNodes, variables)
        : '';
    return `${queryKey}__uiapi__query__${currentFieldName}${argumentString}`;
}
/**
 * fetches a query level ingestion time stamp from the L2 cache
 * if no query has been seen then the timestamp is 0
 * @param apiName
 * @param info
 * @param args
 * @param query
 * @returns
 */
async function fetchIngestionTimeStampFromDatabase(apiName, info, args, query) {
    const { operation, variableValues } = info;
    // if we cannot find the query key in the database then default to 0 as we assume we have not seen the query
    // and all the data is not stale
    let ingestionTimestamp = 0;
    if (info.fieldNodes.length > 0 && info.fieldNodes[0].arguments !== undefined) {
        const key = buildKeyStringForRecordQuery(operation, 
        // join varables passed from query to the argument variables given from the AST
        { ...variableValues, ...args }, info.fieldNodes[0].arguments, apiName);
        return readIngestionTimestampForKey(key, query);
    }
    return ingestionTimestamp;
}

var uiapiSchemaString = "scalar String\nscalar DateTime\nscalar Currency\nscalar ID\nscalar Boolean\nscalar Longitude\nscalar Float\nscalar MultiPicklist\nscalar Base64\nscalar Url\nscalar PhoneNumber\nscalar Email\nscalar TextArea\nscalar Latitude\nscalar Picklist\nscalar RichTextArea\nscalar EncryptedString\nscalar Double\nscalar Long\nscalar JSON\nscalar Time\nscalar Int\nscalar Percent\nscalar LongTextArea\nscalar IdOrRef\nscalar Date\ntype PercentAggregate implements FieldValue {\n  value: Percent\n  displayValue: String\n  avg: DoubleValue\n  count: LongValue\n  countDistinct: LongValue\n  format: String\n  max: PercentValue\n  min: PercentValue\n  sum: PercentValue\n}\n\ntype StringAggregate implements FieldValue {\n  value: String\n  displayValue: String\n  count: LongValue\n  countDistinct: LongValue\n  grouping: IntValue\n  label: String\n  max: StringValue\n  min: StringValue\n}\n\ntype Query {\n  uiapi: UIAPI!\n  setup: Setup__Setup!\n}\n\ninput EmailOperators {\n  eq: Email\n  ne: Email\n  like: Email\n  lt: Email\n  gt: Email\n  lte: Email\n  gte: Email\n  in: [Email]\n  nin: [Email]\n}\n\ninput PolymorphicParentRelationshipRecordOrderBy @generic {\n  RecordOrderBy: RecordOrderBy @fieldCategory\n}\n\ninput DoubleOperators {\n  eq: Double\n  ne: Double\n  lt: Double\n  gt: Double\n  lte: Double\n  gte: Double\n  in: [Double]\n  nin: [Double]\n}\n\ntype DateOnlyAggregation {\n  value: Date\n  format: String\n}\n\ntype RecordCreatePayload @generic {\n  Record: RecordRepresentation\n}\n\ntype DateAggregate implements FieldValue {\n  value: Date\n  displayValue: String\n  calendarMonth: DateFunctionAggregation\n  calendarQuarter: DateFunctionAggregation\n  calendarYear: DateFunctionAggregation\n  count: LongValue\n  countDistinct: LongValue\n  dayInMonth: DateFunctionAggregation\n  dayInWeek: DateFunctionAggregation\n  dayInYear: DateFunctionAggregation\n  fiscalMonth: DateFunctionAggregation\n  fiscalQuarter: DateFunctionAggregation\n  fiscalYear: DateFunctionAggregation\n  format: String\n  grouping: IntValue\n  max: DateValue\n  min: DateValue\n  weekInMonth: DateFunctionAggregation\n  weekInYear: DateFunctionAggregation\n}\n\ninput PolymorphicParentRelationshipGroupBy @generic {\n  RecordGroupBy: RecordGroupBy @fieldCategory\n}\n\nenum GroupByFunction {\n  DAY_IN_WEEK\n  DAY_IN_MONTH\n  DAY_IN_YEAR\n  WEEK_IN_MONTH\n  WEEK_IN_YEAR\n  CALENDAR_MONTH\n  CALENDAR_QUARTER\n  CALENDAR_YEAR\n  FISCAL_MONTH\n  FISCAL_QUARTER\n  FISCAL_YEAR\n  DAY_ONLY\n  HOUR_IN_DAY\n}\n\ntype RecordTypeInfo {\n  available: Boolean!\n  defaultRecordTypeMapping: Boolean!\n  master: Boolean!\n  name: String\n  recordTypeId: ID\n}\n\ninput UIAPIMutationsInput {\n  allOrNone: Boolean = true\n}\n\ntype BooleanValue implements FieldValue {\n  value: Boolean\n  displayValue: String\n}\n\ntype ReferenceToInfo {\n  ApiName: String!\n  nameFields: [String]!\n  objectInfo: ObjectInfo\n}\n\ninterface FieldValue {\n  displayValue: String\n}\n\ntype LongitudeValue implements FieldValue {\n  value: Longitude\n  displayValue: String\n}\n\ntype StringValue implements FieldValue {\n  value: String\n  displayValue: String\n  label: String\n}\n\ntype IntValue implements FieldValue {\n  value: Int\n  displayValue: String\n  format: String\n}\n\ntype UrlValue implements FieldValue {\n  value: Url\n  displayValue: String\n}\n\ninput IdOperators {\n  eq: ID\n  ne: ID\n  lt: ID\n  gt: ID\n  lte: ID\n  gte: ID\n  in: [ID]\n  nin: [ID]\n  inq: JoinInput\n  ninq: JoinInput\n}\n\ninput Setup__SetupOrderBy @generic {\n  orderableField: OrderByClause @fieldCategory\n  orderableGeolocationField: OrderByGeolocationClause @fieldCategory\n  orderableParentRelationship: Setup__SetupOrderBy @fieldCategory\n  orderablePolymorphicParentRelationship: Setup__SetupPolymorphicParentRelationshipRecordOrderBy @fieldCategory\n}\n\ntype LongAggregate implements FieldValue {\n  value: Long\n  displayValue: String\n  avg: DoubleValue\n  count: LongValue\n  countDistinct: LongValue\n  format: String\n  grouping: IntValue\n  max: LongValue\n  min: LongValue\n  sum: LongValue\n}\n\ntype PhoneNumberAggregate implements FieldValue {\n  value: PhoneNumber\n  displayValue: String\n  count: LongValue\n  countDistinct: LongValue\n  grouping: IntValue\n  max: PhoneNumberValue\n  min: PhoneNumberValue\n}\n\ninput TimeOperators {\n  eq: Time\n  ne: Time\n  lt: Time\n  gt: Time\n  lte: Time\n  gte: Time\n  in: [Time]\n  nin: [Time]\n}\n\ntype PicklistValue implements FieldValue {\n  value: Picklist\n  displayValue: String\n  label: String\n}\n\ntype CurrencyAggregate implements FieldValue {\n  value: Currency\n  displayValue: String\n  avg: DoubleValue\n  count: LongValue\n  countDistinct: LongValue\n  format: String\n  max: CurrencyValue\n  min: CurrencyValue\n  sum: CurrencyValue\n}\n\ntype RelatedListInfo {\n  childApiName: String!\n  relatedListName: String!\n  label: String!\n  displayColumns: [ListColumn!]!\n  orderedByInfo: [ListOrder!]!\n  parentApiName: String!\n  fieldApiName: String!\n}\n\ninput StringOperators {\n  eq: String\n  ne: String\n  like: String\n  lt: String\n  gt: String\n  lte: String\n  gte: String\n  in: [String]\n  nin: [String]\n}\n\ntype UIAPI {\n  query: RecordQuery!\n  aggregate: RecordQueryAggregate!\n  objectInfos(apiNames: [String], locale: String): [ObjectInfo]\n  relatedListByName(parentApiName: String!, relatedListName: String!): RelatedListInfo\n}\n\ninput MultiPicklistOperators {\n  eq: MultiPicklist\n  ne: MultiPicklist\n  includes: [MultiPicklist]\n  excludes: [MultiPicklist]\n}\n\ntype DateTimeAggregate implements FieldValue {\n  value: DateTime\n  displayValue: String\n  calendarMonth: DateFunctionAggregation\n  calendarQuarter: DateFunctionAggregation\n  calendarYear: DateFunctionAggregation\n  count: LongValue\n  countDistinct: LongValue\n  dayInMonth: DateFunctionAggregation\n  dayInWeek: DateFunctionAggregation\n  dayInYear: DateFunctionAggregation\n  dayOnly: DateOnlyAggregation\n  fiscalMonth: DateFunctionAggregation\n  fiscalQuarter: DateFunctionAggregation\n  fiscalYear: DateFunctionAggregation\n  format: String\n  hourInDay: DateFunctionAggregation\n  max: DateTimeValue\n  min: DateTimeValue\n  weekInMonth: DateFunctionAggregation\n  weekInYear: DateFunctionAggregation\n}\n\ninput BooleanOperators {\n  eq: Boolean\n  ne: Boolean\n}\n\ntype EmailAggregate implements FieldValue {\n  value: Email\n  displayValue: String\n  count: LongValue\n  countDistinct: LongValue\n  grouping: IntValue\n  max: EmailValue\n  min: EmailValue\n}\n\n#enum OrderByType {\n#}\n\ninput GroupByDateFunction {\n  function: GroupByFunction\n}\n\ntype RichTextAreaValue implements FieldValue {\n  value: RichTextArea\n  displayValue: String\n}\n\ntype MultiPicklistValue implements FieldValue {\n  value: MultiPicklist\n  displayValue: String\n  label: String\n}\n\ntype Setup__SetupEdge @generic {\n  node: Setup__EntityRepresentation\n  cursor: String!\n}\n\ninput DatePrimitiveOperators {\n  eq: Date\n  ne: Date\n  lt: Date\n  gt: Date\n  lte: Date\n  gte: Date\n  in: [Date]\n  nin: [Date]\n}\n\ntype TimeAggregate implements FieldValue {\n  value: Time\n  displayValue: String\n  format: String\n  hourInDay: DateFunctionAggregation\n}\n\ntype __Type {\n  kind: __TypeKind!\n  name: String\n  description: String\n  fields(includeDeprecated: Boolean = false): [__Field!]\n  interfaces: [__Type!]\n  possibleTypes: [__Type!]\n  enumValues(includeDeprecated: Boolean = false): [__EnumValue!]\n  inputFields: [__InputValue!]\n  ofType: __Type\n}\n\ntype ListColumn {\n  fieldApiName: String!\n  label: String!\n  lookupId: String\n  sortable: Boolean\n}\n\ntype Setup__SetupQuery {\n  recordQuery(first: Int, after: String, where: Setup__SetupFilter, orderBy: Setup__SetupOrderBy, scope: String, upperBound: Int): Setup__SetupConnection @fieldCategory\n}\n\ntype Setup__EntityRepresentation @generic {\n  Id: ID!\n  ApiName: String!\n  IntValue: IntValue @fieldCategory\n  StringValue: StringValue @fieldCategory\n  BooleanValue: BooleanValue @fieldCategory\n  IDValue: IDValue @fieldCategory\n  DateTimeValue: DateTimeValue @fieldCategory\n  TimeValue: TimeValue @fieldCategory\n  DateValue: DateValue @fieldCategory\n  TextAreaValue: TextAreaValue @fieldCategory\n  LongTextAreaValue: LongTextAreaValue @fieldCategory\n  RichTextAreaValue: RichTextAreaValue @fieldCategory\n  PhoneNumberValue: PhoneNumberValue @fieldCategory\n  EmailValue: EmailValue @fieldCategory\n  UrlValue: UrlValue @fieldCategory\n  EncryptedStringValue: EncryptedStringValue @fieldCategory\n  CurrencyValue: CurrencyValue @fieldCategory\n  LongitudeValue: LongitudeValue @fieldCategory\n  LatitudeValue: LatitudeValue @fieldCategory\n  PicklistValue: PicklistValue @fieldCategory\n  MultiPicklistValue: MultiPicklistValue @fieldCategory\n  LongValue: LongValue @fieldCategory\n  DoubleValue: DoubleValue @fieldCategory\n  PercentValue: PercentValue @fieldCategory\n  Base64Value: Base64Value @fieldCategory\n  JSONValue: JSONValue @fieldCategory\n  parentRelationship: Setup__EntityRepresentation @fieldCategory\n  polymorphicParentRelationship: Setup__SetupPolymorphicParentRelationship @fieldCategory\n  childRelationship(first: Int, after: String, where: Setup__SetupFilter, orderBy: Setup__SetupOrderBy, upperBound: Int): Setup__SetupConnection @fieldCategory\n  CompoundField: CompoundField @fieldCategory\n}\n\ntype LatitudeAggregate implements FieldValue {\n  value: Latitude\n  displayValue: String\n  avg: DoubleValue\n  count: LongValue\n  countDistinct: LongValue\n  max: LatitudeValue\n  min: LatitudeValue\n  sum: DoubleValue\n}\n\ninput CurrencyOperators {\n  eq: Currency\n  ne: Currency\n  lt: Currency\n  gt: Currency\n  lte: Currency\n  gte: Currency\n  in: [Currency]\n  nin: [Currency]\n}\n\ninput DistanceInput {\n  latitude: Latitude!\n  longitude: Longitude!\n}\n\nunion PolymorphicParentRelationship @generic = RecordRepresentation\n\nenum AggregateOrderByNumberFunction {\n  AVG\n  COUNT\n  COUNT_DISTINCT\n  MAX\n  MIN\n  SUM\n}\n\ntype LongTextAreaValue implements FieldValue {\n  value: LongTextArea\n  displayValue: String\n}\n\ntype LatitudeValue implements FieldValue {\n  value: Latitude\n  displayValue: String\n}\n\ninput OrderByClause {\n  order: ResultOrder\n  nulls: NullOrder\n}\n\ninput AggregateOrderBy @generic {\n  orderableNumberField: AggregateOrderByNumberClause @fieldCategory\n  orderableStringField: AggregateOrderByStringClause @fieldCategory\n  orderableField: NoFunctionAggregateOrderByClause @fieldCategory\n  orderableGeolocationField: OrderByGeolocationClause @fieldCategory\n  orderableParentRelationship: AggregateOrderBy @fieldCategory\n  orderablePolymorphicParentRelationship: PolymorphicParentRelationshipOrderBy @fieldCategory\n  type: String = ORDER_BY\n}\n\ninput GroupByClause {\n  group: Boolean\n}\n\ntype RecordAggregateConnection @generic {\n  edges: [RecordAggregateEdge]\n  pageInfo: PageInfo!\n  totalCount: Int!\n}\n\ntype LongitudeAggregate implements FieldValue {\n  value: Longitude\n  displayValue: String\n  avg: DoubleValue\n  count: LongValue\n  countDistinct: LongValue\n  max: LongitudeValue\n  min: LongitudeValue\n  sum: DoubleValue\n}\n\ntype RecordEdge @generic {\n  node: RecordRepresentation\n  cursor: String!\n}\n\nunion Setup__SetupPolymorphicParentRelationship @generic = Setup__EntityRepresentation\n\ntype DateValue implements FieldValue {\n  value: Date\n  displayValue: String\n  format: String\n}\n\ninput URLOperators {\n  eq: Url\n  ne: Url\n  like: Url\n  lt: Url\n  gt: Url\n  lte: Url\n  gte: Url\n  in: [Url]\n  nin: [Url]\n}\n\ninput LongOperators {\n  eq: Long\n  ne: Long\n  lt: Long\n  gt: Long\n  lte: Long\n  gte: Long\n  in: [Long]\n  nin: [Long]\n}\n\nenum DataType {\n  STRING\n  TEXTAREA\n  PHONE\n  EMAIL\n  URL\n  ENCRYPTEDSTRING\n  BOOLEAN\n  CURRENCY\n  INT\n  LONG\n  DOUBLE\n  PERCENT\n  DATETIME\n  TIME\n  DATE\n  REFERENCE\n  PICKLIST\n  MULTIPICKLIST\n  ADDRESS\n  LOCATION\n  BASE64\n  COMPLEXVALUE\n  COMBOBOX\n  JSON\n  JUNCTIONIDLIST\n  ANYTYPE\n}\n\nenum NullOrder {\n  FIRST\n  LAST\n}\n\ntype PhoneNumberValue implements FieldValue {\n  value: PhoneNumber\n  displayValue: String\n}\n\n# Cannot have empty enum\n# enum RecordScope @generic {\n# }\n\ninput Setup__SetupFilter @generic {\n  and: [Setup__SetupFilter]\n  or: [Setup__SetupFilter]\n  not: Setup__SetupFilter\n  parentRelationshipRecordFilter: Setup__SetupFilter @fieldCategory\n  polymorphicParentRelationshipRecordFilter: Setup__SetupPolymorphicParentRelationshipRecordFilter @fieldCategory\n  IntegerOperator: IntegerOperators @fieldCategory\n  LongOperator: LongOperators @fieldCategory\n  StringOperator: StringOperators @fieldCategory\n  DoubleOperator: DoubleOperators @fieldCategory\n  PercentOperator: PercentOperators @fieldCategory\n  LongitudeOperator: LongitudeOperators @fieldCategory\n  LatitudeOperator: LatitudeOperators @fieldCategory\n  EmailOperator: EmailOperators @fieldCategory\n  TextAreaOperator: TextAreaOperators @fieldCategory\n  LongTextAreaOperator: LongTextAreaOperators @fieldCategory\n  URLOperator: URLOperators @fieldCategory\n  PhoneNumberOperator: PhoneNumberOperators @fieldCategory\n  BooleanOperator: BooleanOperators @fieldCategory\n  Setup__IdOperator: Setup__IdOperators @fieldCategory\n  CurrencyOperator: CurrencyOperators @fieldCategory\n  TimeOperator: TimeOperators @fieldCategory\n  DateOperator: DateOperators @fieldCategory\n  DateTimeOperator: DateTimeOperators @fieldCategory\n  PicklistOperator: PicklistOperators @fieldCategory\n  MultiPicklistOperator: MultiPicklistOperators @fieldCategory\n  GeolocationOperator: GeolocationOperators @fieldCategory\n}\n\ntype DoubleAggregate implements FieldValue {\n  value: Double\n  displayValue: String\n  avg: DoubleValue\n  count: LongValue\n  countDistinct: LongValue\n  format: String\n  max: DoubleValue\n  min: DoubleValue\n  sum: DoubleValue\n}\n\ntype __Field {\n  name: String!\n  description: String\n  args: [__InputValue!]!\n  type: __Type!\n  isDeprecated: Boolean!\n  deprecationReason: String\n}\n\ninput DateOperators {\n  eq: DateInput\n  ne: DateInput\n  lt: DateInput\n  gt: DateInput\n  lte: DateInput\n  gte: DateInput\n  in: [DateInput]\n  nin: [DateInput]\n  DAY_IN_WEEK: DateFunctionInput\n  DAY_IN_MONTH: DateFunctionInput\n  DAY_IN_YEAR: DateFunctionInput\n  WEEK_IN_MONTH: DateFunctionInput\n  WEEK_IN_YEAR: DateFunctionInput\n  CALENDAR_MONTH: DateFunctionInput\n  CALENDAR_QUARTER: DateFunctionInput\n  CALENDAR_YEAR: DateFunctionInput\n  FISCAL_MONTH: DateFunctionInput\n  FISCAL_QUARTER: DateFunctionInput\n  FISCAL_YEAR: DateFunctionInput\n}\n\ninput GeolocationInput {\n  latitude: Latitude!\n  longitude: Longitude!\n  radius: Float!\n  unit: Unit!\n}\n\ninput JoinInput {\n  Record: RecordFilter @fieldCategory\n  ApiName: String\n}\n\ninput TextAreaOperators {\n  eq: TextArea\n  ne: TextArea\n  like: TextArea\n  lt: TextArea\n  gt: TextArea\n  lte: TextArea\n  gte: TextArea\n  in: [TextArea]\n  nin: [TextArea]\n}\n\ntype TextAreaValue implements FieldValue {\n  value: TextArea\n  displayValue: String\n}\n\ntype RecordUpdatePayload @generic {\n  success: Boolean\n}\n\ninput PercentOperators {\n  eq: Percent\n  ne: Percent\n  lt: Percent\n  gt: Percent\n  lte: Percent\n  gte: Percent\n  in: [Percent]\n  nin: [Percent]\n}\n\ninput Setup__SetupPolymorphicParentRelationshipRecordOrderBy @generic {\n  Setup__SetupOrderBy: Setup__SetupOrderBy @fieldCategory\n}\n\ntype DoubleValue implements FieldValue {\n  value: Double\n  displayValue: String\n  format: String\n}\n\ntype IDAggregate implements FieldValue {\n  value: ID\n  displayValue: String\n  count: LongValue\n  countDistinct: LongValue\n  grouping: IntValue\n  max: IDValue\n  min: IDValue\n}\n\ntype __InputValue {\n  name: String!\n  description: String\n  type: __Type!\n  defaultValue: String\n}\n\ntype RecordAggregateEdge @generic {\n  node: RecordResult\n  cursor: String!\n}\n\ntype __Directive {\n  name: String\n  description: String\n  locations: [__DirectiveLocation!]\n  args: [__InputValue!]!\n}\n\ninput RecordCreateInput @generic {\n  record: RecordCreateRepresentation! @fieldCategory\n}\n\ntype ThemeInfo {\n  color: String\n  iconUrl: String\n}\n\ninput AggregateOrderByStringClause {\n  function: AggregateOrderByStringFunction\n  order: ResultsOrder\n  nulls: NullsOrder\n}\n\ntype RecordDeletePayload {\n  Id: ID\n}\n\ntype UrlAggregate implements FieldValue {\n  value: Url\n  displayValue: String\n  count: LongValue\n  countDistinct: LongValue\n  grouping: IntValue\n  max: UrlValue\n  min: UrlValue\n}\n\nenum DateLiteral {\n  LAST_YEAR\n  LAST_WEEK\n  THIS_QUARTER\n  NEXT_FISCAL_YEAR\n  LAST_QUARTER\n  TOMORROW\n  NEXT_FISCAL_QUARTER\n  YESTERDAY\n  NEXT_QUARTER\n  THIS_FISCAL_QUARTER\n  THIS_WEEK\n  LAST_MONTH\n  LAST_90_DAYS\n  NEXT_90_DAYS\n  THIS_FISCAL_YEAR\n  NEXT_WEEK\n  TODAY\n  NEXT_YEAR\n  NEXT_MONTH\n  LAST_FISCAL_QUARTER\n  THIS_MONTH\n  LAST_FISCAL_YEAR\n  THIS_YEAR\n}\n\ntype __EnumValue {\n  name: String!\n  description: String\n  isDeprecated: Boolean!\n  deprecationReason: String\n}\n\ntype RecordRepresentation implements Record  @generic {\n  Id: ID!\n  ApiName: String!\n  WeakEtag: Long!\n  DisplayValue: String\n  LastModifiedById: IDValue\n  LastModifiedDate: DateTimeValue\n  SystemModstamp: DateTimeValue\n  RecordTypeId(fallback: Boolean): IDValue\n  IntValue: IntValue @fieldCategory\n  StringValue: StringValue @fieldCategory\n  BooleanValue: BooleanValue @fieldCategory\n  IDValue: IDValue @fieldCategory\n  DateTimeValue: DateTimeValue @fieldCategory\n  TimeValue: TimeValue @fieldCategory\n  DateValue: DateValue @fieldCategory\n  TextAreaValue: TextAreaValue @fieldCategory\n  LongTextAreaValue: LongTextAreaValue @fieldCategory\n  RichTextAreaValue: RichTextAreaValue @fieldCategory\n  PhoneNumberValue: PhoneNumberValue @fieldCategory\n  EmailValue: EmailValue @fieldCategory\n  UrlValue: UrlValue @fieldCategory\n  EncryptedStringValue: EncryptedStringValue @fieldCategory\n  CurrencyValue: CurrencyValue @fieldCategory\n  LongitudeValue: LongitudeValue @fieldCategory\n  LatitudeValue: LatitudeValue @fieldCategory\n  PicklistValue: PicklistValue @fieldCategory\n  MultiPicklistValue: MultiPicklistValue @fieldCategory\n  LongValue: LongValue @fieldCategory\n  DoubleValue: DoubleValue @fieldCategory\n  PercentValue: PercentValue @fieldCategory\n  Base64Value: Base64Value @fieldCategory\n  JSONValue: JSONValue @fieldCategory\n  parentRelationship: RecordRepresentation @fieldCategory\n  polymorphicParentRelationship: PolymorphicParentRelationship @fieldCategory\n  childRelationship(first: Int, after: String, where: RecordFilter, orderBy: RecordOrderBy, upperBound: Int): RecordConnection @fieldCategory\n  CompoundField: CompoundField @fieldCategory\n}\n\ntype IDValue implements FieldValue {\n  value: ID\n  displayValue: String\n}\n\nenum Unit {\n  MI\n  KM\n}\n\ninput PolymorphicParentRelationshipOrderBy @generic {\n  AggregateOrderBy: AggregateOrderBy @fieldCategory\n}\n\ninput OrderByGeolocationClause {\n  distance: DistanceInput\n  order: ResultOrder\n  nulls: NullOrder\n}\n\ninput Setup__IdOperators {\n  eq: ID\n  ne: ID\n  lt: ID\n  gt: ID\n  lte: ID\n  gte: ID\n  in: [ID]\n  nin: [ID]\n  inq: Setup__JoinInput\n  ninq: Setup__JoinInput\n}\n\nenum NullsOrder {\n  FIRST\n  LAST\n}\n\ntype TextAreaAggregate implements FieldValue {\n  value: TextArea\n  displayValue: String\n  count: LongValue\n  countDistinct: LongValue\n  grouping: IntValue\n  max: TextAreaValue\n  min: TextAreaValue\n}\n\nenum GroupByType {\n  GROUP_BY\n  ROLLUP\n  CUBE\n}\n\nenum ResultOrder {\n  ASC\n  DESC\n}\n\ninput RecordOrderBy @generic {\n  orderableField: OrderByClause @fieldCategory\n  orderableGeolocationField: OrderByGeolocationClause @fieldCategory\n  orderableParentRelationship: RecordOrderBy @fieldCategory\n  orderablePolymorphicParentRelationship: PolymorphicParentRelationshipRecordOrderBy @fieldCategory\n}\n\ninput Setup__JoinInput {\n  Record: Setup__SetupFilter @fieldCategory\n  ApiName: String\n}\n\ninput PicklistOperators {\n  eq: Picklist\n  ne: Picklist\n  in: [Picklist]\n  nin: [Picklist]\n  like: Picklist\n  lt: Picklist\n  gt: Picklist\n  lte: Picklist\n  gte: Picklist\n}\n\nenum ResultsOrder {\n  ASC\n  DESC\n}\n\ninput RecordFilter @generic {\n  and: [RecordFilter]\n  or: [RecordFilter]\n  not: RecordFilter\n  parentRelationshipRecordFilter: RecordFilter @fieldCategory\n  polymorphicParentRelationshipRecordFilter: PolymorphicParentRelationshipRecordFilter @fieldCategory\n  IntegerOperator: IntegerOperators @fieldCategory\n  LongOperator: LongOperators @fieldCategory\n  StringOperator: StringOperators @fieldCategory\n  DoubleOperator: DoubleOperators @fieldCategory\n  PercentOperator: PercentOperators @fieldCategory\n  LongitudeOperator: LongitudeOperators @fieldCategory\n  LatitudeOperator: LatitudeOperators @fieldCategory\n  EmailOperator: EmailOperators @fieldCategory\n  TextAreaOperator: TextAreaOperators @fieldCategory\n  LongTextAreaOperator: LongTextAreaOperators @fieldCategory\n  URLOperator: URLOperators @fieldCategory\n  PhoneNumberOperator: PhoneNumberOperators @fieldCategory\n  BooleanOperator: BooleanOperators @fieldCategory\n  IdOperator: IdOperators @fieldCategory\n  CurrencyOperator: CurrencyOperators @fieldCategory\n  TimeOperator: TimeOperators @fieldCategory\n  DateOperator: DateOperators @fieldCategory\n  DateTimeOperator: DateTimeOperators @fieldCategory\n  PicklistOperator: PicklistOperators @fieldCategory\n  MultiPicklistOperator: MultiPicklistOperators @fieldCategory\n  GeolocationOperator: GeolocationOperators @fieldCategory\n}\n\ntype TimeValue implements FieldValue {\n  value: Time\n  displayValue: String\n  format: String\n}\n\ninput GeolocationOperators {\n  lt: GeolocationInput\n  gt: GeolocationInput\n}\n\ntype PicklistAggregate implements FieldValue {\n  value: Picklist\n  displayValue: String\n  count: LongValue\n  countDistinct: LongValue\n  grouping: IntValue\n  label: String\n  max: PicklistValue\n  min: PicklistValue\n}\n\ninput LatitudeOperators {\n  eq: Latitude\n  ne: Latitude\n  lt: Latitude\n  gt: Latitude\n  lte: Latitude\n  gte: Latitude\n  in: [Latitude]\n  nin: [Latitude]\n}\n\ninput RecordUpdateRepresentation @generic {\n  Int: Int @fieldCategory\n  String: String @fieldCategory\n  Boolean: Boolean @fieldCategory\n  ID: IdOrRef @fieldCategory\n  DateTime: DateTime @fieldCategory\n  Time: Time @fieldCategory\n  Date: Date @fieldCategory\n  TextArea: TextArea @fieldCategory\n  LongTextArea: LongTextArea @fieldCategory\n  RichTextArea: RichTextArea @fieldCategory\n  PhoneNumber: PhoneNumber @fieldCategory\n  Email: Email @fieldCategory\n  Url: Url @fieldCategory\n  EncryptedString: EncryptedString @fieldCategory\n  Currency: Currency @fieldCategory\n  Longitude: Longitude @fieldCategory\n  Latitude: Latitude @fieldCategory\n  Picklist: Picklist @fieldCategory\n  MultiPicklist: MultiPicklist @fieldCategory\n  Long: Long @fieldCategory\n  Double: Double @fieldCategory\n  Percent: Percent @fieldCategory\n  Base64: Base64 @fieldCategory\n  JSON: JSON @fieldCategory\n}\n\ntype DateTimeValue implements FieldValue {\n  value: DateTime\n  displayValue: String\n  format: String\n}\n\ninput RecordDeleteInput {\n  Id: IdOrRef!\n}\n\nenum __DirectiveLocation {\n  QUERY\n  MUTATION\n  FIELD\n  FRAGMENT_DEFINITION\n  FRAGMENT_SPREAD\n  INLINE_FRAGMENT\n  SCHEMA\n  SCALAR\n  OBJECT\n  FIELD_DEFINITION\n  ARGUMENT_DEFINITION\n  INTERFACE\n  UNION\n  ENUM\n  ENUM_VALUE\n  INPUT_OBJECT\n  INPUT_FIELD_DEFINITION\n}\n\ntype IntAggregate implements FieldValue {\n  value: Int\n  displayValue: String\n  avg: DoubleValue\n  count: LongValue\n  countDistinct: LongValue\n  format: String\n  grouping: IntValue\n  max: IntValue\n  min: IntValue\n  sum: LongValue\n}\n\ntype ListOrder {\n  fieldApiName: String!\n  sortDirection: ResultOrder\n}\n\ntype RecordAggregate @generic {\n  ApiName: String!\n  BooleanAggregate: BooleanAggregate @fieldCategory\n  CurrencyAggregate: CurrencyAggregate @fieldCategory\n  DateAggregate: DateAggregate @fieldCategory\n  DoubleAggregate: DoubleAggregate @fieldCategory\n  EmailAggregate: EmailAggregate @fieldCategory\n  IDAggregate: IDAggregate @fieldCategory\n  IntAggregate: IntAggregate @fieldCategory\n  LatitudeAggregate: LatitudeAggregate @fieldCategory\n  LongitudeAggregate: LongitudeAggregate @fieldCategory\n  LongAggregate: LongAggregate @fieldCategory\n  PercentAggregate: PercentAggregate @fieldCategory\n  PhoneNumberAggregate: PhoneNumberAggregate @fieldCategory\n  PicklistAggregate: PicklistAggregate @fieldCategory\n  StringAggregate: StringAggregate @fieldCategory\n  TextAreaAggregate: TextAreaAggregate @fieldCategory\n  TimeAggregate: TimeAggregate @fieldCategory\n  UrlAggregate: UrlAggregate @fieldCategory\n}\n\ntype JSONValue implements FieldValue {\n  value: JSON\n  displayValue: String\n}\n\ntype EmailValue implements FieldValue {\n  value: Email\n  displayValue: String\n}\n\ntype Setup__Setup {\n  query: Setup__SetupQuery!\n}\n\nenum AggregateOrderByStringFunction {\n  COUNT\n  COUNT_DISTINCT\n  MAX\n  MIN\n}\n\ntype LongValue implements FieldValue {\n  value: Long\n  displayValue: String\n  format: String\n}\n\ninput DateFunctionInput {\n  value: LongOperators\n  convertTimezoneValue: LongOperators\n}\n\n# Mutations aren't supported yet.\n#type Mutation {\n#  uiapi(input: UIAPIMutationsInput): UIAPIMutations!\n#}\n\ntype DependentField {\n  controllingField: String!\n  dependentFields: [String]!\n}\n\ninput LongTextAreaOperators {\n  eq: LongTextArea\n  ne: LongTextArea\n  like: LongTextArea\n  lt: LongTextArea\n  gt: LongTextArea\n  lte: LongTextArea\n  gte: LongTextArea\n  in: [LongTextArea]\n  nin: [LongTextArea]\n}\n\nenum __TypeKind {\n  SCALAR\n  OBJECT\n  INTERFACE\n  UNION\n  ENUM\n  INPUT_OBJECT\n  LIST\n  NON_NULL\n}\n\ntype Setup__SetupConnection @generic {\n  edges: [Setup__SetupEdge]\n  pageInfo: PageInfo!\n  totalCount: Int!\n  pageResultCount: Int!\n}\n\ntype PercentValue implements FieldValue {\n  value: Percent\n  displayValue: String\n  format: String\n}\n\ninput DateTimeOperators {\n  eq: DateTimeInput\n  ne: DateTimeInput\n  lt: DateTimeInput\n  gt: DateTimeInput\n  lte: DateTimeInput\n  gte: DateTimeInput\n  in: [DateTimeInput]\n  nin: [DateTimeInput]\n  DAY_IN_WEEK: DateFunctionInput\n  DAY_IN_MONTH: DateFunctionInput\n  DAY_IN_YEAR: DateFunctionInput\n  WEEK_IN_MONTH: DateFunctionInput\n  WEEK_IN_YEAR: DateFunctionInput\n  CALENDAR_MONTH: DateFunctionInput\n  CALENDAR_QUARTER: DateFunctionInput\n  CALENDAR_YEAR: DateFunctionInput\n  FISCAL_MONTH: DateFunctionInput\n  FISCAL_QUARTER: DateFunctionInput\n  FISCAL_YEAR: DateFunctionInput\n  DAY_ONLY: DateTimeFunctionInput\n  HOUR_IN_DAY: DateFunctionInput\n}\n\ninput NoFunctionAggregateOrderByClause {\n  order: ResultsOrder\n  nulls: NullsOrder\n}\n\ntype BooleanAggregate implements FieldValue {\n  value: Boolean\n  displayValue: String\n  grouping: IntValue\n}\n\ntype RecordQueryAggregate {\n    # RecordScope is replaced with String\n  recordQueryAggregate(first: Int, after: String, where: RecordFilter, orderBy: AggregateOrderBy, scope: String, groupBy: RecordGroupBy, upperBound: Int): RecordAggregateConnection @fieldCategory\n}\n\ntype RecordConnection @generic {\n  edges: [RecordEdge]\n  pageInfo: PageInfo!\n  totalCount: Int!\n  pageResultCount: Int!\n}\n\ntype FilteredLookupInfo {\n  controllingFields: [String]!\n  dependent: Boolean!\n  optionalFilter: Boolean!\n}\n\ninput PhoneNumberOperators {\n  eq: PhoneNumber\n  ne: PhoneNumber\n  like: PhoneNumber\n  lt: PhoneNumber\n  gt: PhoneNumber\n  lte: PhoneNumber\n  gte: PhoneNumber\n  in: [PhoneNumber]\n  nin: [PhoneNumber]\n}\n\ntype ObjectInfo {\n  ApiName: String!\n  childRelationships: [ChildRelationship]!\n  createable: Boolean!\n  custom: Boolean!\n  defaultRecordTypeId: ID\n  deletable: Boolean!\n  dependentFields: [DependentField]!\n  feedEnabled: Boolean!\n  fields: [Field]!\n  keyPrefix: String\n  label: String\n  labelPlural: String\n  layoutable: Boolean!\n  mruEnabled: Boolean!\n  nameFields: [String]!\n  queryable: Boolean!\n  recordTypeInfos: [RecordTypeInfo]!\n  searchable: Boolean!\n  themeInfo: ThemeInfo\n  updateable: Boolean!\n  locale: String\n}\n\ninput LongitudeOperators {\n  eq: Longitude\n  ne: Longitude\n  lt: Longitude\n  gt: Longitude\n  lte: Longitude\n  gte: Longitude\n  in: [Longitude]\n  nin: [Longitude]\n}\n\ninput RecordCreateRepresentation @generic {\n  Int: Int @fieldCategory\n  String: String @fieldCategory\n  Boolean: Boolean @fieldCategory\n  ID: IdOrRef @fieldCategory\n  DateTime: DateTime @fieldCategory\n  Time: Time @fieldCategory\n  Date: Date @fieldCategory\n  TextArea: TextArea @fieldCategory\n  LongTextArea: LongTextArea @fieldCategory\n  RichTextArea: RichTextArea @fieldCategory\n  PhoneNumber: PhoneNumber @fieldCategory\n  Email: Email @fieldCategory\n  Url: Url @fieldCategory\n  EncryptedString: EncryptedString @fieldCategory\n  Currency: Currency @fieldCategory\n  Longitude: Longitude @fieldCategory\n  Latitude: Latitude @fieldCategory\n  Picklist: Picklist @fieldCategory\n  MultiPicklist: MultiPicklist @fieldCategory\n  Long: Long @fieldCategory\n  Double: Double @fieldCategory\n  Percent: Percent @fieldCategory\n  Base64: Base64 @fieldCategory\n  JSON: JSON @fieldCategory\n}\n\ntype Field {\n  ApiName: String!\n  calculated: Boolean!\n  compound: Boolean!\n  compoundComponentName: String\n  compoundFieldName: String\n  controllerName: String\n  controllingFields: [String]!\n  createable: Boolean!\n  custom: Boolean!\n  dataType: DataType\n  extraTypeInfo: FieldExtraTypeInfo\n  filterable: Boolean!\n  filteredLookupInfo: FilteredLookupInfo\n  highScaleNumber: Boolean!\n  htmlFormatted: Boolean!\n  inlineHelpText: String\n  label: String\n  nameField: Boolean!\n  polymorphicForeignKey: Boolean!\n  precision: Int\n  reference: Boolean!\n  referenceTargetField: String\n  referenceToInfos: [ReferenceToInfo]!\n  relationshipName: String\n  required: Boolean!\n  scale: Int\n  searchPrefilterable: Boolean\n  sortable: Boolean!\n  updateable: Boolean!\n}\n\nenum FieldExtraTypeInfo {\n  IMAGE_URL\n  EXTERNAL_LOOKUP\n  INDIRECT_LOOKUP\n  PERSONNAME\n  SWITCHABLE_PERSONNAME\n  PLAINTEXTAREA\n  RICHTEXTAREA\n}\n\ntype RateLimit {\n  cost: Long\n  limit: Long\n  remaining: Long\n  resetAt: DateTime\n}\n\ninput DateRange {\n  last_n_days: Int\n  next_n_days: Int\n  last_n_weeks: Int\n  next_n_weeks: Int\n  last_n_months: Int\n  next_n_months: Int\n  last_n_quarters: Int\n  next_n_quarters: Int\n  last_n_fiscal_quarters: Int\n  next_n_fiscal_quarters: Int\n  last_n_years: Int\n  next_n_years: Int\n  last_n_fiscal_years: Int\n  next_n_fiscal_years: Int\n  n_days_ago: Int\n  n_weeks_ago: Int\n  n_months_ago: Int\n  n_quarters_ago: Int\n  n_years_ago: Int\n  n_fiscal_quarters_ago: Int\n  n_fiscal_years_ago: Int\n}\n\ntype UIAPIMutations {\n  recordCreate(input: RecordCreateInput!): RecordCreatePayload @fieldCategory\n  recordDelete(input: RecordDeleteInput!): RecordDeletePayload @fieldCategory\n  recordUpdate(input: RecordUpdateInput!): RecordUpdatePayload @fieldCategory\n}\n\ninput DateTimeFunctionInput {\n  value: DatePrimitiveOperators\n  convertTimezoneValue: DatePrimitiveOperators\n}\n\ntype Base64Value implements FieldValue {\n  value: Base64\n  displayValue: String\n}\n\ninput IntegerOperators {\n  eq: Int\n  ne: Int\n  lt: Int\n  gt: Int\n  lte: Int\n  gte: Int\n  in: [Int]\n  nin: [Int]\n}\n\ntype EncryptedStringValue implements FieldValue {\n  value: EncryptedString\n  displayValue: String\n}\n\ninterface Record {\n  Id: ID!\n  ApiName: String!\n  WeakEtag: Long!\n  DisplayValue: String\n  LastModifiedById: IDValue\n  LastModifiedDate: DateTimeValue\n  SystemModstamp: DateTimeValue\n  RecordTypeId(fallback: Boolean): IDValue\n}\n\ninput PolymorphicParentRelationshipRecordFilter @generic {\n  RecordFilter: RecordFilter @fieldCategory\n}\n\ninput AggregateOrderByNumberClause {\n  function: AggregateOrderByNumberFunction\n  order: ResultsOrder\n  nulls: NullsOrder\n}\n\ntype __Schema {\n  types: [__Type!]!\n  queryType: __Type!\n  mutationType: __Type\n  directives: [__Directive!]!\n  subscriptionType: __Type\n}\n\ninput Setup__SetupPolymorphicParentRelationshipRecordFilter @generic {\n  Setup__SetupFilter: Setup__SetupFilter @fieldCategory\n}\n\ntype CompoundField @generic {\n  IntValue: IntValue @fieldCategory\n  StringValue: StringValue @fieldCategory\n  BooleanValue: BooleanValue @fieldCategory\n  IDValue: IDValue @fieldCategory\n  DateTimeValue: DateTimeValue @fieldCategory\n  TimeValue: TimeValue @fieldCategory\n  DateValue: DateValue @fieldCategory\n  TextAreaValue: TextAreaValue @fieldCategory\n  LongTextAreaValue: LongTextAreaValue @fieldCategory\n  RichTextAreaValue: RichTextAreaValue @fieldCategory\n  PhoneNumberValue: PhoneNumberValue @fieldCategory\n  EmailValue: EmailValue @fieldCategory\n  UrlValue: UrlValue @fieldCategory\n  EncryptedStringValue: EncryptedStringValue @fieldCategory\n  CurrencyValue: CurrencyValue @fieldCategory\n  LongitudeValue: LongitudeValue @fieldCategory\n  LatitudeValue: LatitudeValue @fieldCategory\n  PicklistValue: PicklistValue @fieldCategory\n  MultiPicklistValue: MultiPicklistValue @fieldCategory\n  LongValue: LongValue @fieldCategory\n  DoubleValue: DoubleValue @fieldCategory\n  PercentValue: PercentValue @fieldCategory\n  Base64Value: Base64Value @fieldCategory\n  JSONValue: JSONValue @fieldCategory\n}\n\ninput RecordUpdateInput @generic {\n  Id: IdOrRef!\n  record: RecordUpdateRepresentation! @fieldCategory\n}\n\ninput DateTimeInput {\n  value: DateTime\n  literal: DateLiteral\n  range: DateRange\n}\n\ntype ChildRelationship {\n  childObjectApiName: String!\n  fieldName: String\n  junctionIdListNames: [String]!\n  junctionReferenceTo: [String]!\n  relationshipName: String\n  objectInfo: ObjectInfo\n}\n\ntype RecordResult @generic {\n  aggregate: RecordAggregate\n}\n\ntype PageInfo {\n  hasNextPage: Boolean!\n  hasPreviousPage: Boolean!\n  startCursor: String\n  endCursor: String\n}\n\ntype CurrencyValue implements FieldValue {\n  value: Currency\n  displayValue: String\n  format: String\n}\n\ninput DateInput {\n  value: Date\n  literal: DateLiteral\n  range: DateRange\n}\n\ninput RecordGroupBy @generic {\n  groupableField: GroupByClause @fieldCategory\n  groupableDateField: GroupByDateFunction @fieldCategory\n  groupableParentRelationship: RecordGroupBy @fieldCategory\n  groupablePolymorphicParentRelationship: PolymorphicParentRelationshipGroupBy @fieldCategory\n  type: GroupByType = GROUP_BY\n}\n\ntype DateFunctionAggregation {\n  value: Long\n  format: String\n}\n\ntype RecordQuery {\n    # scope should be type RecordScope but that's empty enum.\n  recordQuery(first: Int, after: String, where: RecordFilter, orderBy: RecordOrderBy, scope: String, upperBound: Int): RecordConnection @fieldCategory\n}\n\ndirective @generic on OBJECT | INTERFACE | UNION | ENUM | INPUT_OBJECT\ndirective @fieldCategory on FIELD_DEFINITION | ARGUMENT_DEFINITION | INPUT_FIELD_DEFINITION | ENUM_VALUE\ndirective @category(name: String!) on FIELD";

// Define additional schema that is missing from uiapi that we use in local evaluation
const additionalSchemaDefinitions = /* GraphQL */ `
    enum SupportedScopes {
        ASSIGNEDTOME
        MINE
    }
`;
const baseTypeDefinitions = uiapiSchemaString + additionalSchemaDefinitions;
class CachedGraphQLSchema {
    constructor() {
        this._schema = buildBaseSchema();
        this._polymorphicFieldTypeNames = [];
    }
    getSchema() {
        return this._schema;
    }
    setSchema(value) {
        this._schema = value;
    }
    getPolymorphicFieldTypeNames() {
        return this._polymorphicFieldTypeNames;
    }
    setPolymorphicFieldTypeNames(value) {
        this._polymorphicFieldTypeNames = value;
    }
    set(schema, polymorphicFieldTypeNames) {
        this._schema = schema;
        this._polymorphicFieldTypeNames = polymorphicFieldTypeNames;
    }
}
/**
 * Builds the base schema from uiapi graphql adapter with resolvers attached
 * @returns GraphQLSchema
 */
function buildBaseSchema() {
    return addResolversToSchema(buildSchema(baseTypeDefinitions), []);
}

({
    kind: Kind.DIRECTIVE,
    name: {
        kind: Kind.NAME,
        value: 'category',
    },
    arguments: [
        {
            kind: Kind.ARGUMENT,
            name: {
                kind: Kind.NAME,
                value: 'name',
            },
            value: {
                kind: Kind.STRING,
                value: PARENT_RELATIONSHIP,
                block: false,
            },
        },
    ],
});
({
    kind: Kind.SELECTION_SET,
    selections: [
        {
            kind: Kind.FIELD,
            name: {
                kind: Kind.NAME,
                value: 'value',
            },
        },
    ],
});

/**
 * Copyright (c) 2022, Salesforce, Inc.,
 * All rights reserved.
 * For full license text, see the LICENSE.txt file
 */


const { keys, values, create, assign, freeze } = Object;

function createLink(key) {
    return { __ref: key };
}
/**
 * Records are stored in the durable store with scalar fields denormalized. This function takes that denoramlized
 * durable store record representation and normalizes it back out into the format the the luvio store expects it
 * @param key Record store key
 * @param entry Durable entry containing a denormalized record representation
 * @returns a set of entries containing the normalized record and its normalized fields
 */
function normalizeRecordFields(key, entry) {
    const { data: record } = entry;
    const { fields, links } = record;
    const missingFieldLinks = keys(links);
    const fieldNames = keys(fields);
    const normalizedFields = {};
    const returnEntries = {};
    // restore fields
    for (let i = 0, len = fieldNames.length; i < len; i++) {
        const fieldName = fieldNames[i];
        const field = fields[fieldName];
        const fieldKey = buildRecordFieldStoreKey(key, fieldName);
        returnEntries[fieldKey] = { data: field };
        normalizedFields[fieldName] = createLink(fieldKey);
    }
    // restore missing fields
    for (let i = 0, len = missingFieldLinks.length; i < len; i++) {
        const fieldName = missingFieldLinks[i];
        const link = links[fieldName];
        if (link.isMissing === true) {
            normalizedFields[fieldName] = { ...link, __ref: undefined };
        }
    }
    returnEntries[key] = {
        data: assign(record, { fields: normalizedFields }),
        metadata: entry.metadata,
    };
    return returnEntries;
}
/**
 * Transforms a record for storage in the durable store. The transformation involves denormalizing
 * scalar fields and persisting link metadata to transform back into a normalized representation
 *
 * If the record contains pending fields this will return undefined as pending records do not get persisted
 * to the durable store. There should be a refresh operation outbound that will bring in the updated record.
 *
 * @param normalizedRecord Record containing normalized field links
 * @param recordStore a store containing referenced record fields
 */
function buildDurableRecordRepresentation(normalizedRecord, records, pendingEntries, store) {
    const fields = normalizedRecord.fields;
    const filteredFields = {};
    const links = {};
    const fieldNames = keys(fields);
    for (let i = 0, len = fieldNames.length; i < len; i++) {
        const fieldName = fieldNames[i];
        const field = fields[fieldName];
        // pending fields get filtered out of the durable store
        const { pending } = field;
        if (pending === true) {
            // do not write records with pending fields to the durable store
            // there should be a refresh operation outbound that will bring in the updated record
            return undefined;
        }
        const { __ref } = field;
        if (__ref !== undefined) {
            let ref = records[__ref];
            if (pendingEntries !== undefined) {
                // If the ref was part of the pending write that takes precedence
                const pendingEntry = pendingEntries[__ref];
                if (pendingEntry !== undefined) {
                    ref = pendingEntry.data;
                }
            }
            // if field reference exists then add it to our filteredFields
            if (ref !== undefined) {
                filteredFields[fieldName] = ref;
            }
            else {
                // if we have a store to read, try to find the field there too
                // The durable ingest staging store may pass through to L1, and
                // not all fields are necessarily published every time, so it is
                // important to check L1 and not just the fields being published,
                // otherwise we risk truncating the fields on the record.
                if (store) {
                    ref = store.readEntry(__ref);
                    if (ref !== undefined) {
                        filteredFields[fieldName] = ref;
                    }
                }
            }
        }
        // we want to preserve fields that are missing nodes
        if (field.isMissing === true) {
            links[fieldName] = field;
        }
    }
    return {
        ...normalizedRecord,
        fields: filteredFields,
        links,
    };
}
function getDenormalizedKey(originalKey, recordId, luvio) {
    // this will likely need to be handled when moving to structured keys
    // note record view entities dont have an associated keybuilder. They get ingested as records to a different key format
    // see the override for how they are handled packages/lds-adapters-uiapi/src/raml-artifacts/types/RecordRepresentation/keyBuilderFromType.ts
    if (originalKey.startsWith(RECORD_VIEW_ENTITY_ID_PREFIX)) {
        return RECORD_VIEW_ENTITY_ID_PREFIX + recordId;
    }
    return keyBuilderRecord(luvio, { recordId });
}
function makeRecordDenormalizingDurableStore(luvio, durableStore, getStoreRecords, getStoreMetadata, getStore, sqlStore) {
    const getEntries = function (entries, segment) {
        // this HOF only inspects records in the default segment
        if (segment !== DefaultDurableSegment) {
            return durableStore.getEntries(entries, segment);
        }
        const { length: entriesLength } = entries;
        if (entriesLength === 0) {
            return Promise.resolve({});
        }
        // filter out record field keys
        const filteredEntryIds = [];
        // map of records to avoid requesting duplicate record keys when requesting both records and fields
        const recordEntries = {};
        const recordViewEntries = {};
        for (let i = 0, len = entriesLength; i < len; i++) {
            const id = entries[i];
            const recordId = extractRecordIdFromStoreKey(id);
            if (recordId !== undefined) {
                if (id.startsWith(RECORD_VIEW_ENTITY_ID_PREFIX)) {
                    if (recordViewEntries[recordId] === undefined) {
                        const key = getDenormalizedKey(id, recordId, luvio);
                        recordViewEntries[recordId] = true;
                        filteredEntryIds.push(key);
                    }
                }
                else {
                    if (recordEntries[recordId] === undefined) {
                        const key = getDenormalizedKey(id, recordId, luvio);
                        recordEntries[recordId] = true;
                        filteredEntryIds.push(key);
                    }
                }
            }
            else {
                filteredEntryIds.push(id);
            }
        }
        // call base getEntries
        return durableStore.getEntries(filteredEntryIds, segment).then((durableEntries) => {
            if (durableEntries === undefined) {
                return undefined;
            }
            const returnEntries = create(null);
            const keys$1 = keys(durableEntries);
            for (let i = 0, len = keys$1.length; i < len; i++) {
                const key = keys$1[i];
                const value = durableEntries[key];
                if (value === undefined) {
                    continue;
                }
                if (isEntryDurableRecordRepresentation(value, key)) {
                    assign(returnEntries, normalizeRecordFields(key, value));
                }
                else {
                    returnEntries[key] = value;
                }
            }
            return returnEntries;
        });
    };
    const denormalizeEntries = function (entries) {
        let hasEntries = false;
        let hasMetadata = false;
        const putEntries = create(null);
        const putMetadata = create(null);
        const keys$1 = keys(entries);
        const putRecords = {};
        const putRecordViews = {};
        const storeRecords = getStoreRecords !== undefined ? getStoreRecords() : {};
        const storeMetadata = getStoreMetadata !== undefined ? getStoreMetadata() : {};
        const store = getStore();
        for (let i = 0, len = keys$1.length; i < len; i++) {
            const key = keys$1[i];
            let value = entries[key];
            const recordId = extractRecordIdFromStoreKey(key);
            // do not put normalized field values
            if (recordId !== undefined) {
                const isRecordView = key.startsWith(RECORD_VIEW_ENTITY_ID_PREFIX);
                if (isRecordView) {
                    if (putRecordViews[recordId] === true) {
                        continue;
                    }
                }
                else {
                    if (putRecords[recordId] === true) {
                        continue;
                    }
                }
                const recordKey = getDenormalizedKey(key, recordId, luvio);
                const recordEntries = entries;
                const entry = recordEntries[recordKey];
                let record = entry && entry.data;
                if (record === undefined) {
                    record = storeRecords[recordKey];
                    if (record === undefined) {
                        // fields are being published without a record for them existing,
                        // fields cannot exist standalone in the durable store
                        continue;
                    }
                }
                if (isRecordView) {
                    putRecordViews[recordId] = true;
                }
                else {
                    putRecords[recordId] = true;
                }
                if (isStoreRecordError(record)) {
                    hasEntries = true;
                    putEntries[recordKey] = value;
                    continue;
                }
                let metadata = entry && entry.metadata;
                if (metadata === undefined) {
                    metadata = {
                        ...storeMetadata[recordKey],
                        metadataVersion: DURABLE_METADATA_VERSION,
                    };
                }
                const denormalizedRecord = buildDurableRecordRepresentation(record, storeRecords, recordEntries, store);
                if (denormalizedRecord !== undefined) {
                    hasEntries = true;
                    putEntries[recordKey] = {
                        data: denormalizedRecord,
                        metadata,
                    };
                    // if undefined then it is pending
                    // we should still update metadata on pending records
                }
                else {
                    hasMetadata = true;
                    metadata.expirationTimestamp = metadata.ingestionTimestamp;
                    putMetadata[recordKey] = {
                        metadata,
                    };
                }
            }
            else {
                hasEntries = true;
                putEntries[key] = value;
            }
        }
        return { putEntries, putMetadata, hasEntries, hasMetadata };
    };
    const setEntries = function (entries, segment) {
        if (segment !== DefaultDurableSegment) {
            return durableStore.setEntries(entries, segment);
        }
        const { putEntries, putMetadata, hasEntries, hasMetadata } = denormalizeEntries(entries);
        const promises = [
            hasEntries ? durableStore.setEntries(putEntries, segment) : undefined,
        ];
        if (sqlStore !== undefined && sqlStore.isBatchUpdateSupported()) {
            promises.push(hasMetadata && sqlStore !== undefined
                ? durableStore.setMetadata(putMetadata, segment)
                : undefined);
        }
        return Promise.all(promises).then(() => { });
    };
    const batchOperations = function (operations) {
        const operationsWithDenormedRecords = [];
        for (let i = 0, len = operations.length; i < len; i++) {
            const operation = operations[i];
            if (operation.type === 'setMetadata') {
                // if setMetadata also contains entry data then it needs to be denormalized.
                const keys$1 = keys(operation.entries);
                if (keys$1.length > 0) {
                    const firstKey = keys$1[0];
                    // casted to any to check if data exists
                    const firstEntry = operation.entries[firstKey];
                    // it is not possible for setMetadata to contain entries with both data and no data in the same operation.
                    // this is determined by the plugin supporting update batch calls before it gets to this HOF.
                    // so we only need to check one entry to confirm this for performance
                    if (firstEntry.data !== undefined) {
                        const { putEntries, putMetadata, hasMetadata } = denormalizeEntries(operation.entries);
                        operationsWithDenormedRecords.push({
                            ...operation,
                            entries: putEntries,
                        });
                        if (hasMetadata &&
                            sqlStore !== undefined &&
                            sqlStore.isBatchUpdateSupported() === true) {
                            operationsWithDenormedRecords.push({
                                ...operation,
                                entries: putMetadata,
                                type: 'setMetadata',
                            });
                        }
                    }
                    else {
                        operationsWithDenormedRecords.push(operation);
                    }
                }
                continue;
            }
            if (operation.segment !== DefaultDurableSegment || operation.type === 'evictEntries') {
                operationsWithDenormedRecords.push(operation);
                continue;
            }
            const { putEntries, putMetadata, hasMetadata } = denormalizeEntries(operation.entries);
            operationsWithDenormedRecords.push({
                ...operation,
                entries: putEntries,
            });
            if (hasMetadata &&
                sqlStore !== undefined &&
                sqlStore.isBatchUpdateSupported() === true) {
                operationsWithDenormedRecords.push({
                    ...operation,
                    entries: putMetadata,
                    type: 'setMetadata',
                });
            }
        }
        return durableStore.batchOperations(operationsWithDenormedRecords);
    };
    /**
     * Retrieves a denormalized record from the store
     * NOTE: do no use this if you don't know what you're doing, this can still contain normalized record references
     * @param recordKey record key
     * @param durableStore the durable store
     * @returns a DraftRecordRepresentation containing the requested fields
     */
    const getDenormalizedRecord = function (recordKey) {
        return durableStore.getEntries([recordKey], DefaultDurableSegment).then((entries) => {
            if (entries === undefined) {
                return undefined;
            }
            const denormalizedEntry = entries[recordKey];
            if (denormalizedEntry === undefined) {
                return undefined;
            }
            // don't include link information
            const denormalizedRecord = denormalizedEntry.data;
            if (isStoreRecordError(denormalizedRecord)) {
                return undefined;
            }
            return denormalizedRecord;
        });
    };
    return create(durableStore, {
        getEntries: { value: getEntries, writable: true },
        setEntries: { value: setEntries, writable: true },
        batchOperations: { value: batchOperations, writable: true },
        getDenormalizedRecord: { value: getDenormalizedRecord, writable: true },
    });
}
// create the runtime cache for the graphql schema when the factory creates the adapter
new CachedGraphQLSchema();

let luvio;
let getIngestRecords;
let getIngestMetadata;
let getIngestStore;
// LdsSqliteStore plugin helper
function getNimbusDurableStore() {
    const resolvedPlugin = __nimbus.plugins.LdsSqliteStore;
    return new NimbusSqliteStore(resolvedPlugin);
}
/**
 * Use record denormalizing store to persist field values and references
 * When not used, subsequent reads may result in empty field values
 * See TD-0153300 for moving RecordDenormalizingDurableStore out of lds-drafts
 */
function createRecordDenormingStore(luvio, durableStore) {
    const recordDenormingStore = makeRecordDenormalizingDurableStore(luvio, durableStore, () => getIngestRecords(), () => getIngestMetadata(), () => getIngestStore());
    return recordDenormingStore;
}
function getRuntime() {
    // Create stores
    const store = new InMemoryStore();
    const durableStore = getNimbusDurableStore();
    const recordDenormingStore = createRecordDenormingStore(luvio, durableStore);
    // Create environment with store + aura network adapter
    const baseEnv = new Environment(store, networkAdapter);
    const durableEnv = makeDurable(baseEnv, {
        durableStore: recordDenormingStore,
    });
    // Set ingest records/metadata properties from durable environment
    getIngestRecords = durableEnv.getIngestStagingStoreRecords;
    getIngestMetadata = durableEnv.getIngestStagingStoreMetadata;
    getIngestStore = durableEnv.getIngestStagingStore;
    // Return new luvio instance
    luvio = new Luvio(durableEnv, {
        instrument: instrumentLuvio,
    });
    return luvio;
}

// so eslint doesn't complain about nimbus
/* global __nimbus */
function ldsRuntimeBridge() {
    if (typeof __nimbus !== 'undefined' &&
        __nimbus.plugins !== undefined &&
        __nimbus.plugins.LdsSqliteStore !== undefined) {
        // Since LdsSqliteStore plugin present use Luvio with persistent store
        const luvio = getRuntime();
        setDefaultLuvio({ luvio });
        return { name: 'ldsRuntimeBridge' };
    }
    else {
        // Plugin not available fallback to lds-runtime-aura/ldsEngineCreator
        return ldsEngineCreator();
    }
}

export { ldsRuntimeBridge as default };
// version: 1.287.0-dev10-0c7922050
